'''Script to fit RGB heatmap images to a source color palette.'''
import click
import numpy as np
from PIL import Image

from zennit.image import CMAPS, palette


def gale_shapley(dist):
    '''Find a stable matching given a distance matrix.'''
    preference = np.argsort(dist, axis=1)
    proposed = np.zeros(dist.shape[0], dtype=int)
    loners = set(range(dist.shape[0]))
    guys = [-1] * dist.shape[0]
    gals = [-1] * dist.shape[1]
    while loners:
        loner = loners.pop()
        target = preference[loner, proposed[loner]]
        if gals[target] == -1:
            gals[target] = loner
            guys[loner] = target
        elif dist[gals[target], target] > dist[loner, target]:
            gals[target] = loner
            guys[loner] = target
            guys[gals[target]] = -1
            loners.add(gals[target])
        else:
            loners.add(loner)
        proposed[loner] += 1
    return guys


@click.command()
@click.argument('source-file', type=click.Path(exists=True, dir_okay=False))
@click.argument('output-file', type=click.Path(writable=True, dir_okay=False))
@click.option('--strategy', type=click.Choice(['intensity', 'nearest', 'histogram']), default='intensity')
@click.option('--source-cmap', type=click.Choice(list(CMAPS)), default='bwr')
@click.option('--source-level', type=float, default=1.0)
@click.option('--invert/--no-invert', default=False)
@click.option('--cmap', type=click.Choice(list(CMAPS)), default='coldnhot')
@click.option('--level', type=float, default=1.0)
def main(source_file, output_file, strategy, source_cmap, source_level, invert, cmap, level):
    '''Fit an existing RGB heatmap image to a color palette.'''
    source = np.array(Image.open(source_file).convert('RGB'))
    matchpal = palette(source_cmap, source_level)

    if strategy == 'intensity':
        # matching based on the source image intensity/ brightness
        values = source.astype(float).mean(2)
    elif strategy == 'nearest':
        # match by finding the neareast centroids in a source colormap
        dists = (np.abs(source[None].astype(float) - matchpal[:, None, None].astype(float))).sum(3)
        values = np.argmin(dists, axis=0)
    elif strategy == 'histogram':
        # match by finding a stable match between the color histogram of the source image and a source colormap
        source = np.concatenate([source, np.zeros_like(source[:, :, [0]])], axis=2).view(np.uint32)[..., 0]
        uniques, counts = np.unique(source, return_counts=True)
        uniques = uniques[np.argsort(counts)[-256:]]
        dist = (np.abs(uniques.view(np.uint8).reshape(-1, 1, 4)[..., :3] - matchpal[None])).sum(2)
        matches = np.array(gale_shapley(dist))

        ind_bin, ind_h, ind_w = np.nonzero(source[None] == uniques[:, None, None])
        values = np.zeros(source.shape[:2], dtype=np.uint8)
        values[ind_h, ind_w] = matches[ind_bin]

    values = values.clip(0, 255).astype(np.uint8)
    if invert:
        values = 255 - values

    img = Image.fromarray(values, mode='P')
    pal = palette(cmap, level)
    img.putpalette(pal)
    img.save(output_file)


if __name__ == '__main__':
    main()



'''Script to swap the palette of heatmap images.'''
import click
from PIL import Image

from zennit.image import CMAPS, palette


@click.command()
@click.argument('image-files', type=click.Path(exists=True, dir_okay=False), nargs=-1)
@click.option('--cmap', type=click.Choice(list(CMAPS)), default='coldnhot')
@click.option('--level', type=float, default=1.0)
def main(image_files, cmap, level):
    '''Swap the palette of heatmap image files inline.'''
    for fname in image_files:
        img = Image.open(fname)
        img = img.convert('P')
        pal = palette(cmap, level)
        img.putpalette(pal)
        img.save(fname)


if __name__ == '__main__':
    main()



'''Tests for image operations.'''
from typing import NamedTuple
from itertools import product
from io import BytesIO

import pytest
import numpy as np
from PIL import Image

from zennit.cmap import ColorMap
from zennit.image import get_cmap, palette, imgify, gridify, imsave, interval_norm_bounds


@pytest.fixture(scope='session', params=[
    'gray', '000,fff', ColorMap('000,fff')
])
def cmap_source(request):
    '''Fixture for multiple ways to specify the "gray" color map.'''
    return request.param


class ImageTuple(NamedTuple):
    '''NamedTuple for image-array setups.'''
    grid: bool
    nchannels: list
    channel_front: bool
    width: int
    height: int
    array: np.ndarray


@pytest.fixture(scope='session', params=product(
    [False, True],
    [1, 3],
    [False, True],
    [5, 10],
    [5, 10],
    [np.float64, np.uint8]
))
def image_tuple(request):
    '''Image-array setups with varying size, type, number of channels, channel position and grid dimension.'''
    grid, nchannels, channel_front, width, height, dtype = request.param

    shape = (height, width)
    if channel_front:
        shape = (nchannels,) + shape
    else:
        shape = shape + (nchannels,)
    shape = (1,) * grid + shape

    return ImageTuple(
        grid,
        nchannels,
        channel_front,
        width,
        height,
        np.ones(shape, dtype=dtype),
    )


def test_get_cmap(cmap_source):
    '''Test whether get_cmap handles its supported cmap types correctly.'''
    cmap = get_cmap(cmap_source)
    assert isinstance(cmap, ColorMap), 'Returned object is not a ColorMap!'
    assert cmap.source == '000,fff', 'Mismatch in source code of returned ColorMap instance.'


def test_palette(cmap_source):
    '''Test whether palette returns the correct palette for all of its supported types.'''
    pal = palette(cmap_source)
    expected_pal = np.repeat(np.arange(256, dtype=np.uint8)[:, None], 3, axis=1)
    assert np.allclose(expected_pal, pal)


@pytest.mark.parametrize('ndim', [1, 4, 5, 6])
def test_imgify_wrong_dim(ndim):
    '''Test whether imgify fails for an unsupported number of dimensions.'''
    with pytest.raises(TypeError):
        imgify(np.zeros((1,) * ndim))


@pytest.mark.parametrize('ndim', [1, 2, 5, 6])
def test_imgify_grid_wrong_dim(ndim):
    '''Test whether imgify fails for an unsupported number of dimensions with grid=True.'''
    with pytest.raises(TypeError):
        imgify(np.zeros((1,) * ndim), grid=True)


@pytest.mark.parametrize('grid', [[1], (1,), 1, [1, 1, 1], (1, 1, 1)])
def test_imgify_grid_bad_grid(grid):
    '''Test whether imgify fails for unsupported grid values.'''
    with pytest.raises(TypeError):
        imgify(np.zeros((1,) * 4), grid=grid)


@pytest.mark.parametrize('grid,nchannels', product([False, True], [2, 4]))
def test_imgify_wrong_channels(grid, nchannels):
    '''Test whether imgify fails for an unsupported number of dimensions with grid=True.'''
    with pytest.raises(TypeError):
        imgify(np.zeros((1,) * grid + (2, 2, nchannels)), grid=grid)


def test_imgify_container(image_tuple):
    '''Test whether imgify produces the correct PIL Image container'''
    image = imgify(image_tuple.array, grid=image_tuple.grid)
    assert image.mode == ('P' if image_tuple.nchannels == 1 else 'RGB'), 'Mode mismatch!'
    assert image.width == image_tuple.width, 'Width mismatch!'
    assert image.height == image_tuple.height, 'Height mismatch!'


@pytest.mark.parametrize('vmin,vmax,symmetric', product([None, 1.], [None, 2.], [False, True]))
def test_imgify_normalization(vmin, vmax, symmetric):
    '''Test whether imgify normalizes as expected.'''
    array = np.array([[-1., 0., 3.]])

    image = imgify(array, cmap='gray', vmin=vmin, vmax=vmax, symmetric=symmetric)

    if vmin is None:
        if symmetric:
            vmin = -np.abs(array).max()
        else:
            vmin = array.min()
    if vmax is None:
        if symmetric:
            vmax = np.abs(array).max()
        else:
            vmax = array.max()

    expected = (((array - vmin) / (vmax - vmin)) * 255.).clip(0, 255).astype(np.uint8)

    assert np.allclose(np.array(image), expected)


@pytest.mark.parametrize('ndim', [1, 2, 5, 6])
def test_gridify_wrong_dim(ndim):
    '''Test whether imgify fails for an unsupported number of dimensions.'''
    with pytest.raises(TypeError):
        gridify(np.zeros((1,) * ndim))


@pytest.mark.parametrize('channel_front,nchannels', product([False, True], [2, 4]))
def test_gridify_wrong_channels(channel_front, nchannels):
    '''Test whether gridify fails for an unsupported number of channels in both channel positions.'''
    shape = (2, 2)
    if channel_front:
        shape = (nchannels,) + shape
    else:
        shape = shape + (nchannels,)
    shape = (1,) + shape

    with pytest.raises(TypeError):
        gridify(np.zeros(shape))


@pytest.mark.parametrize('shape,expected_shape', [
    [(4, 2, 2, 3), (4, 4, 3)],
    [(4, 2, 2, 1), (4, 4, 1)],
    [(4, 2, 2), (4, 4, 1)],
    [(4, 3, 2, 2), (4, 4, 3)],
    [(4, 1, 2, 2), (4, 4, 1)],
])
def test_gridify_shape(shape, expected_shape):
    '''Test whether gridify produces the correct shape.'''
    output = gridify(np.zeros(shape))
    assert expected_shape == output.shape


@pytest.mark.parametrize('fill_value', [None, 0.])
def test_gridify_fill(fill_value):
    '''Test whether gridify fills empty pixels with the correct value.'''
    array = np.array([[[[1.]]]])
    output = gridify(array, fill_value=fill_value, shape=(1, 2))
    expected_value = array.min() if fill_value is None else fill_value
    assert output[0, 1, 0] == expected_value


@pytest.mark.parametrize('writer_params', [None, {}])
def test_imsave_container(image_tuple, writer_params):
    '''Test whether imsave produces a file, which loads as the correct PIL Image container.'''
    fp = BytesIO()
    imsave(fp, image_tuple.array, grid=image_tuple.grid, format='png', writer_params=writer_params)
    fp.seek(0)
    image = Image.open(fp)
    assert image.mode == ('P' if image_tuple.nchannels == 1 else 'RGB'), 'Mode mismatch!'
    assert image.width == image_tuple.width, 'Width mismatch!'
    assert image.height == image_tuple.height, 'Height mismatch!'


@pytest.mark.parametrize('symmetric,dim,expected_bounds', [
    (False, None, (np.array([[[[-1.]]], [[[0.]]]]), np.array([[[[-0.2]]], [[[0.8]]]]))),
    (False, (1, 2, 3), (np.array([[[[-1.]]], [[[0.]]]]), np.array([[[[-0.2]]], [[[0.8]]]]))),
    (False, (0, 1, 2, 3), (np.array([[[[-1.]]]]), np.array([[[[0.8]]]]))),
    (True, None, (np.array([[[[-1.]]], [[[-0.8]]]]), np.array([[[[1.]]], [[[0.8]]]]))),
    (True, (1, 2, 3), (np.array([[[[-1.]]], [[[-0.8]]]]), np.array([[[[1.]]], [[[0.8]]]]))),
    (True, (0, 1, 2, 3), (np.array([[[[-1.]]]]), np.array([[[[1.]]]]))),
])
def test_interval_norm_bounds(symmetric, dim, expected_bounds):
    '''Test whether interval_norm_bounds computes the correct minimum and maximum values.'''
    array = np.linspace(-1., 0.8, 10).reshape((2, 1, 5, 1))
    bounds = interval_norm_bounds(array, symmetric=symmetric, dim=dim)
    assert np.allclose(expected_bounds, bounds)



'''Configuration and fixtures for testing'''
import random
from itertools import product, groupby
from collections import OrderedDict

import pytest
import torch
from torch.nn import Conv1d, ConvTranspose1d, Linear
from torch.nn import Conv2d, ConvTranspose2d
from torch.nn import Conv3d, ConvTranspose3d
from torch.nn import BatchNorm1d, BatchNorm2d, BatchNorm3d
from torchvision.models import vgg11, resnet18, alexnet
from helpers import prodict, one_hot_max

from zennit.attribution import identity
from zennit.core import Composite, Hook
from zennit.composites import COMPOSITES
from zennit.composites import EpsilonGammaBox
from zennit.composites import LayerMapComposite
from zennit.composites import MixedComposite
from zennit.composites import NameLayerMapComposite
from zennit.composites import NameMapComposite
from zennit.composites import SpecialFirstLayerMapComposite
from zennit.types import Linear as AnyLinear, Activation


def pytest_addoption(parser):
    '''Add options to pytest.'''
    parser.addoption(
        '--batchsize',
        default=4,
        help='Batch-size for generated samples.'
    )


def pytest_generate_tests(metafunc):
    '''Generate test fixture values based on CLI options.'''
    if 'batchsize' in metafunc.fixturenames:
        metafunc.parametrize('batchsize', [metafunc.config.getoption('batchsize')], scope='session')


@pytest.fixture(
    scope='session',
    params=[
        0xdeadbeef,
        0xd0c0ffee,
        *[pytest.param(seed, marks=pytest.mark.extended) for seed in [
            0xc001bee5, 0xc01dfee7, 0xbe577001, 0xca7b0075, 0x1057b0a7, 0x900ddeed
        ]],
    ],
    ids=hex
)
def rng(request):
    '''Fixture for the NumPy random number generator.'''
    return torch.manual_seed(request.param)


@pytest.fixture(scope='session')
def pyrng(rng):
    '''Fixture for the Python random number generator.'''
    return random.Random(rng.initial_seed())


@pytest.fixture(
    scope='session',
    params=[
        (torch.nn.ReLU, {}),
        (torch.nn.Softmax, {'dim': 1}),
        (torch.nn.Tanh, {}),
        (torch.nn.Sigmoid, {}),
        (torch.nn.Softplus, {'beta': 1}),
    ],
    ids=lambda param: param[0].__name__
)
def module_simple(rng, request):
    '''Fixture for simple modules.'''
    module_type, kwargs = request.param
    return module_type(**kwargs).to(torch.float64).eval()


@pytest.fixture(
    scope='session',
    params=[
        *product(
            [Linear],
            prodict(in_features=[16], out_features=[15], bias=[True, False]),
        ),
        *product(
            [Conv1d, Conv2d, Conv3d, ConvTranspose1d, ConvTranspose2d, ConvTranspose3d],
            prodict(in_channels=[1, 3], out_channels=[1, 3], kernel_size=[2, 3], bias=[True, False]),
        ),
    ],
    ids=lambda param: param[0].__name__
)
def module_linear(rng, request):
    '''Fixture for linear modules.'''
    module_type, kwargs = request.param
    return module_type(**kwargs).to(torch.float64).eval()


@pytest.fixture(scope='session')
def module_batchnorm(module_linear, rng):
    '''Fixture for BatchNorm-type modules, based on adjacent linear module.'''
    module_map = [
        ((Linear, Conv1d, ConvTranspose1d), BatchNorm1d),
        ((Conv2d, ConvTranspose2d), BatchNorm2d),
        ((Conv3d, ConvTranspose3d), BatchNorm3d),
    ]
    feature_index_map = [
        ((ConvTranspose1d, ConvTranspose2d, ConvTranspose3d), 1),
        ((Linear, Conv1d, Conv2d, Conv3d), 0),
    ]

    batchnorm_type = None
    for types, target_type in module_map:
        if isinstance(module_linear, types):
            batchnorm_type = target_type
            break
    if batchnorm_type is None:
        raise RuntimeError('No batchnorm type for linear layer found.')

    feature_index = None
    for types, index in feature_index_map:
        if isinstance(module_linear, types):
            feature_index = index
            break
    if feature_index is None:
        raise RuntimeError('No feature index for linear layer found.')

    batchnorm = batchnorm_type(num_features=module_linear.weight.shape[feature_index]).to(torch.float64).eval()
    batchnorm.weight.data.uniform_(**{'from': 0.1, 'to': 2.0, 'generator': rng})
    batchnorm.bias.data.normal_(generator=rng)
    batchnorm.eps = 1e-30
    return batchnorm


@pytest.fixture(scope='session')
def data_linear(rng, batchsize, module_linear):
    '''Fixture to create data for a linear module, given an RNG.'''
    shape = (batchsize,)
    setups = [
        (Conv1d, 1, 1),
        (ConvTranspose1d, 0, 1),
        (Conv2d, 1, 2),
        (ConvTranspose2d, 0, 2),
        (Conv3d, 1, 3),
        (ConvTranspose3d, 0, 3)
    ]
    if isinstance(module_linear, Linear):
        shape += (module_linear.weight.shape[1],)
    else:
        for module_type, dim, ndims in setups:
            if isinstance(module_linear, module_type):
                shape += (module_linear.weight.shape[dim],) + (4,) * ndims

    return torch.empty(*shape, dtype=torch.float64).normal_(generator=rng)


@pytest.fixture(scope='session', params=[
    (16,),
    (4,),
    (4, 4),
    (4, 4, 4),
])
def data_simple(request, rng, batchsize):
    '''Fixture to create data for a linear module, given an RNG.'''
    shape = (batchsize,) + request.param
    return torch.empty(*shape, dtype=torch.float64).normal_(generator=rng)


COMPOSITE_KWARGS = {
    EpsilonGammaBox: {'low': -3., 'high': 3.},
}


class PassClone(Hook):
    '''Clone of the Pass rule.'''
    def backward(self, module, grad_input, grad_output):
        '''Directly return grad_output.'''
        return grad_output


class GradClone(Hook):
    '''Explicit rule to return the cloned gradient.'''
    def backward(self, module, grad_input, grad_output):
        '''Directly return grad_output.'''
        return grad_input.clone()


@pytest.fixture(scope='session', params=[
    None,
    [(Linear, GradClone()), (Activation, PassClone())],
])
def cooperative_layer_map(request):
    '''Fixture for a cooperative layer map in LayerMapComposite subtypes.'''
    return request.param


@pytest.fixture(scope='session', params=[
    None,
    [(AnyLinear, GradClone())],
])
def cooperative_first_map(request):
    '''Fixture for a cooperative layer map for the first layer in SpecialFirstLayerMapComposite subtypes.'''
    return request.param


@pytest.fixture(scope='session', params=[
    elem for elem in COMPOSITES.values()
    if issubclass(elem, LayerMapComposite) and not issubclass(elem, SpecialFirstLayerMapComposite)
])
def layer_map_composite(request, cooperative_layer_map):
    '''Fixture for explicit LayerMapComposites.'''
    return request.param(layer_map=cooperative_layer_map, **COMPOSITE_KWARGS.get(request.param, {}))


@pytest.fixture(scope='session', params=[
    elem for elem in COMPOSITES.values() if issubclass(elem, SpecialFirstLayerMapComposite)
])
def special_first_layer_map_composite(request, cooperative_layer_map, cooperative_first_map):
    '''Fixturer for explicit SpecialFirstLayerMapComposites.'''
    return request.param(
        layer_map=cooperative_layer_map,
        first_map=cooperative_first_map,
        **COMPOSITE_KWARGS.get(request.param, {})
    )


@pytest.fixture(scope='session', params=[Composite, *COMPOSITES.values()])
def any_composite(request):
    '''Fixture for all explicitly registered Composites, as well as the empty Composite.'''
    return request.param(**COMPOSITE_KWARGS.get(request.param, {}))


@pytest.fixture(scope='session')
def name_map_composite(model_vision, layer_map_composite):
    '''Fixture to create NameMapComposites based on explicit LayerMapComposites.'''
    rule_map = {}
    for name, child in model_vision.named_modules():
        for dtype, hook_template in layer_map_composite.layer_map:
            if isinstance(child, dtype):
                rule_map.setdefault(hook_template, []).append(name)
                break
    name_map = [(tuple(value), key) for key, value in rule_map.items()]
    return NameMapComposite(name_map=name_map)


@pytest.fixture(scope='session')
def partial_name_map_composite(name_map_composite, pyrng):
    '''Fixture to create a randomly sampled partial NameMapComposites.'''
    name_map = name_map_composite.name_map
    assocs = [(i, j) for i, (keys, _) in enumerate(name_map) for j in range(len(keys))]
    accepted_assocs = sorted(pyrng.sample(assocs, len(assocs) // 2))
    partial_name_map = [
        (tuple(name_map[k][0][n] for _, n in g), name_map[k][1].copy())
        for k, g in groupby(accepted_assocs, lambda o: o[0])
    ]

    return NameMapComposite(name_map=partial_name_map)


@pytest.fixture(scope='session')
def mixed_composite(partial_name_map_composite, special_first_layer_map_composite):
    '''Fixture to create NameLayerMapComposites based on an explicit NameMapComposite and
    SpecialFirstLayerMapComposites.
    '''
    composites = [partial_name_map_composite, special_first_layer_map_composite]
    return MixedComposite(composites)


@pytest.fixture(scope='session')
def name_layer_map_composite(partial_name_map_composite, layer_map_composite):
    '''Fixture to create NameLayerMapComposites based on an explicit NameMapComposite and LayerMapComposite.'''
    return NameLayerMapComposite(
        name_map=partial_name_map_composite.name_map,
        layer_map=layer_map_composite.layer_map,
    )


@pytest.fixture(scope='session', params=[alexnet, vgg11, resnet18])
def model_vision(request):
    '''Models to test composites on.'''
    return request.param()


@pytest.fixture(scope='session')
def model_simple(rng, module_linear, data_linear):
    '''Fixture for a simple model, using a linear module followed by a ReLU and a dense layer.'''
    with torch.no_grad():
        intermediate = module_linear(data_linear)
    return torch.nn.Sequential(OrderedDict([
        ('linr0', module_linear),
        ('actv0', torch.nn.ReLU()),
        ('flat0', torch.nn.Flatten()),
        ('linr1', torch.nn.Linear(intermediate.shape[1:].numel(), 4, dtype=intermediate.dtype)),
    ]))


@pytest.fixture(scope='session')
def model_simple_grad(data_linear, model_simple):
    '''Fixture for gradient wrt. data_linear for model_simple.'''
    data = data_linear.detach().requires_grad_()
    output = model_simple(data)
    grad, = torch.autograd.grad(output, data, output)
    return grad


@pytest.fixture(scope='session')
def model_simple_output(data_linear, model_simple):
    '''Fixture for output given data_linear for model_simple.'''
    data = data_linear.detach()
    output = model_simple(data)
    return output


@pytest.fixture(scope='session', params=[
    identity,
    one_hot_max,
    torch.ones_like,
])
def grad_outputs_func(request):
    '''Fixture for common attr_output_fn functions.'''
    return request.param



'''Script to visually inspect color maps.'''
import click
import numpy as np
from PIL import Image

from zennit.image import CMAPS, palette


def semsstr(string):
    if isinstance(string, list):
        return string
    return [obj for obj in string.split(';') if obj]


@click.command()
@click.argument('output')
@click.option('--cmap', 'colormap_src', type=semsstr, default=list(CMAPS))
@click.option('--level', type=float, default=1.0)
def main(output, colormap_src, level):
    print('\n'.join(colormap_src))
    palettes = np.stack([palette(obj, level) for obj in colormap_src])
    arr = np.repeat(palettes, 32, 0)
    img = Image.fromarray(arr)
    img.save(output)


if __name__ == '__main__':
    main()



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/layer.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Additional Utility Layers'''
import torch


class Sum(torch.nn.Module):
    '''Compute the sum along an axis.

    Parameters
    ----------
    dim : int
        Dimension over which to sum.
    '''
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, input):
        '''Computes the sum along a dimension.'''
        return torch.sum(input, dim=self.dim)



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/attribution.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Attributors are convenience objects to compute attributions, optionally using composites.'''
from abc import ABCMeta, abstractmethod
from functools import partial
from itertools import product

import torch


def constant(obj):
    '''Wrapper function to create a function which returns a constant object regardless of arguments.

    Parameters
    ----------
    obj : object
        Constant object which the returned wrapper function will return on call.

    Returns
    -------
    wrapped_const : function
        Function which when called with any arguments will return ``obj``.

    '''
    def wrapped_const(*args, **kwargs):
        return obj
    return wrapped_const


def identity(obj):
    '''Identity function.

    Parameters
    ----------
    obj : object
        Any object which will be returned.

    Result
    ------
    obj : object
        The original input argument ``obj``.

    '''
    return obj


def occlude_independent(input, mask, fill_fn=torch.zeros_like, invert=False):
    '''Given a ``mask``, occlude pixels of ``input`` independently given a function ``fill_fn``.

    Parameters
    ----------
    input : :py:obj:`torch.Tensor`
        The input tensor which will be occluded in the pixels where ``mask`` is non-zero, or, if ``invert`` is
        ``True``, where ``mask`` is zero, using function ``fill_fn``.
    mask : :py:obj:`torch.Tensor`
        Boolean mask, at which non-zero or zero (given ``invert``) elements will be occluded in ``input``.
    fill_fn : function, optional
        Function used to occlude pixels with signature ``(input : torch.Tensor) -> torch.Tensor``, where input is the
        same shape as ``input``, and the output shall leave the shape unchanged. Default is ``torch.zeros_like``, which
        will replace occluded pixels with zero.
    invert : bool, optional
        If ``True``, inverts the supplied mask. Default is ``False``, i.e. not to invert.

    Returns
    -------
    :py:obj:`torch.Tensor`
        The occluded tensor.

    '''
    if invert:
        mask = ~mask
    return input * mask + ~mask * fill_fn(input)


class Attributor(metaclass=ABCMeta):
    '''Base Attributor Class.

    Attributors are convenience objects with an optional composite and when called, compute an attribution, e.g., the
    gradient or anything that is the result of computing the gradient when using the provided composite.  Attributors
    also provide a context to be used in a `with` statement, similar to `CompositeContext`s. If the forward function
    (or `self.__call__`) is called and the composite has not been registered (i.e. `composite.handles` is empty), the
    composite will be temporarily registered to the model.

    Parameters
    ----------
    model: :py:obj:`torch.nn.Module`
        The model for which the attribution will be computed. If `composite` is provided, this will also be the model
        to which the composite will be registered within `with` statements, or when calling the `Attributor` instance.
    composite: :py:obj:`zennit.core.Composite`, optional
        The optional composite to, if provided, be registered to the model within `with` statements or when calling the
        `Attributor` instance.
    attr_output: :py:obj:`torch.Tensor` or callable, optional
        The default output attribution to be used when calling the `Attributor` instance, which is either a Tensor
        compatible with any input used, or a function of the model's output. If None (default), the value will be the
        identity function.

    '''
    def __init__(self, model, composite=None, attr_output=None):
        self.model = model
        self.composite = composite

        if attr_output is None:
            self.attr_output_fn = identity
        elif not callable(attr_output):
            self.attr_output_fn = constant(attr_output)
        else:
            self.attr_output_fn = attr_output

    def __enter__(self):
        '''Register the composite, if provided.

        Returns
        -------
        self: :py:obj:`Attributor`
            The `Attributor` instance.

        '''
        if self.composite is not None:
            self.composite.register(self.model)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''Remove the composite, if provided.

        Returns
        -------
        False

        '''
        if self.composite is not None:
            self.composite.remove()
        return False

    def __call__(self, input, attr_output=None):
        '''Compute the attribution of the model wrt. `input`, using `attr_output` as the output attribution if
        provided, or the default output attribution otherwise (if not supplied during instantiation either, this will
        be the full output of the model). If a composite was supplied to the `Attributor` instance, but it was not yet
        registered (either manually, or in a `with` statement), it will be registered to the model temporarily during
        the call.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution, which is either a Tensor compatible `input` (i.e. has the same shape as the output
            of the model), or a function of the model's output. If None (default), the default attribution will be
            used, which if neither supplied, will result in the model output used as the output attribution.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model with argument `input`.
        attribution: :py:obj:`torch.Tensor`
            Attribution of the model wrt. to `input`, with the same shape as `input`.
        '''
        if attr_output is None:
            attr_output_fn = self.attr_output_fn
        elif not callable(attr_output):
            attr_output_fn = constant(attr_output)
        else:
            attr_output_fn = attr_output

        if self.composite is None or self.composite.handles:
            return self.forward(input, attr_output_fn)

        with self:
            return self.forward(input, attr_output_fn)

    @property
    def inactive(self):
        '''Return the attributor's composite's ``.inactive`` context.'''
        return self.composite.inactive

    @abstractmethod
    def forward(self, input, attr_output_fn):
        '''Abstract method. Compute the attribution of the model wrt. input, by using `attr_output_fn` as the function
        of the model output to provide the output attribution. This function will not register the composite, and is
        wrapped in the `__call__` of `Attributor`.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution function of the model's output.
        '''


class Gradient(Attributor):
    '''The Gradient Attributor. The result is the product of the attribution output and the (possibly modified)
    jacobian. With a composite, i.e. `EpsilonGammaBox`, this will compute the Layer-wise Relevance Propagation
    attribution values.

    Parameters
    ----------
    model: :py:obj:`torch.nn.Module`
        The model for which the attribution will be computed. If `composite` is provided, this will also be the model
        to which the composite will be registered within `with` statements, or when calling the `Attributor` instance.
    composite: :py:obj:`zennit.core.Composite`, optional
        The optional composite to, if provided, be registered to the model within `with` statements or when calling the
        `Attributor` instance.
    attr_output: :py:obj:`torch.Tensor` or callable, optional
        The default output attribution to be used when calling the `Attributor` instance, which is either a Tensor
        compatible with any input used, or a function of the model's output. If None (default), the value will be the
        identity function.
    create_graph: bool, optional
        Specify whether to use ``create_graph=True`` (default is False) to compute the gradient with
        :py:obj:`torch.autograd.grad`. This needs to be `True` to compute higher order gradients.
    retain_graph: bool, optional
        Specify whether to use ``retain_graph=True`` (default is the value of create_graph) to compute the gradient
        with :py:obj:`torch.autograd.grad`.
    '''
    def __init__(self, model, composite=None, attr_output=None, create_graph=False, retain_graph=None):
        super().__init__(model=model, composite=composite, attr_output=attr_output)
        self.create_graph = create_graph
        self.retain_graph = retain_graph

    def grad(self, input, attr_output_fn):
        '''Compute the gradient of the model wrt. input, by using ``attr_output_fn`` as the function of the model
        output to provide the vector for the vector jacobian product.
        This function is used by subclasses to compute the gradient with the same parameters.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model.
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution function of the model's output.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model given ``input``.
        gradient: :py:obj:`torch.Tensor`
            Gradient of the model wrt. to ``input``, with the same shape as ``input``.
        '''
        if not input.requires_grad:
            input.requires_grad = True
        output = self.model(input)
        gradient, = torch.autograd.grad(
            (output,),
            (input,),
            grad_outputs=(attr_output_fn(output),),
            create_graph=self.create_graph,
            retain_graph=self.retain_graph,
        )
        return output, gradient

    def forward(self, input, attr_output_fn):
        '''Compute the gradient of the model wrt. input, by using `attr_output_fn` as the function of the model output
        to provide the vector for the vector jacobian product.
        This function will not register the composite, and is wrapped in the `__call__` of `Attributor`.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution.
        attr_output_fn: :py:obj:`torch.Tensor` or callable
            The output attribution function of the model's output.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model given `input`.
        attribution: :py:obj:`torch.Tensor`
            Attribution of the model wrt. to `input`, with the same shape as `input`.
        '''
        # create a view of input in case it does not already requires grad
        input = input.view_as(input)
        return self.grad(input, attr_output_fn)


class SmoothGrad(Gradient):
    '''This implements SmoothGrad [1]_. The result is the average over the gradient of multiple iterations where some
    normal distributed noise was added to the input. Supplying a composite will result instead in averaging over the
    modified gradient.

    Parameters
    ----------
    model: :py:obj:`torch.nn.Module`
        The model for which the attribution will be computed. If `composite` is provided, this will also be the model
        to which the composite will be registered within `with` statements, or when calling the `Attributor` instance.
    composite: :py:obj:`zennit.core.Composite`, optional
        The optional composite to, if provided, be registered to the model within `with` statements or when calling the
        `Attributor` instance.
    attr_output: :py:obj:`torch.Tensor` or callable, optional
        The default output attribution to be used when calling the `Attributor` instance, which is either a Tensor
        compatible with any input used, or a function of the model's output. If None (default), the value will be the
        identity function.
    create_graph: bool, optional
        Specify whether to use ``create_graph=True`` (default is False) to compute the gradient with
        :py:obj:`torch.autograd.grad`. This needs to be `True` to compute higher order gradients.
    retain_graph: bool, optional
        Specify whether to use ``retain_graph=True`` (default is the value of create_graph) to compute the gradient
        with :py:obj:`torch.autograd.grad`.
    noise_level: float, optional
        The noise level, which is :math:`\\frac{\\sigma}{x_{max} - x_{min}}` and defaults to 0.1.
    n_iter: int, optional
        The number of iterations over which to average, defaults to 20.

    References
    ----------
    .. [1] D. Smilkov, N. Thorat, B. Kim, F. B. Viégas, and M. Wattenberg: "SmoothGrad: removing noise by adding
           noise," CoRR, vol. abs/1706.03825, 2017.

    '''
    def __init__(
        self, model, composite=None, attr_output=None, create_graph=False, retain_graph=None, noise_level=0.1, n_iter=20
    ):
        super().__init__(
            model=model,
            composite=composite,
            attr_output=attr_output,
            create_graph=create_graph,
            retain_graph=retain_graph
        )
        self.noise_level = noise_level
        self.n_iter = n_iter

    def forward(self, input, attr_output_fn):
        '''Compute the SmoothGrad of the model wrt. input, by using `attr_output_fn` as the function of the model output
        to provide the vector for the vector jacobian product used to compute the gradient.
        This function will not register the composite, and is wrapped in the `__call__` of `Attributor`.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution.
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution function of the model's output.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model given `input`.
        attribution: :py:obj:`torch.Tensor`
            Attribution of the model wrt. to `input`, with the same shape as `input`.
        '''
        dims = tuple(range(1, input.ndim))
        std = self.noise_level * (input.amax(dims, keepdim=True) - input.amin(dims, keepdim=True))

        result = torch.zeros_like(input)
        for n in range(self.n_iter):
            # the last epsilon is defined as zero to compute the true output,
            # and have SmoothGrad w/ n_iter = 1 === gradient
            if n == self.n_iter - 1:
                epsilon = torch.zeros_like(input)
            else:
                epsilon = torch.randn_like(input) * std
            output, gradient = self.grad(input + epsilon, attr_output_fn)
            result += gradient / self.n_iter

        # output is leaking from the loop for the last epsilon (which is zero)
        return output, result


class IntegratedGradients(Gradient):
    '''This implements Integrated Gradients [2]_. The result is the path integral of the gradients, estimated over
    multiple discrete iterations. Supplying a composite will result instead in the path integral over the modified
    gradient.

    Parameters
    ----------
    model: :py:obj:`torch.nn.Module`
        The model for which the attribution will be computed. If `composite` is provided, this will also be the model
        to which the composite will be registered within `with` statements, or when calling the `Attributor` instance.
    composite: :py:obj:`zennit.core.Composite`, optional
        The optional composite to, if provided, be registered to the model within `with` statements or when calling the
        `Attributor` instance.
    attr_output: :py:obj:`torch.Tensor` or callable, optional
        The default output attribution to be used when calling the `Attributor` instance, which is either a Tensor
        compatible with any input used, or a function of the model's output. If None (default), the value will be the
        identity function.
    create_graph: bool, optional
        Specify whether to use ``create_graph=True`` (default is False) to compute the gradient with
        :py:obj:`torch.autograd.grad`. This needs to be `True` to compute higher order gradients.
    retain_graph: bool, optional
        Specify whether to use ``retain_graph=True`` (default is the value of create_graph) to compute the gradient
        with :py:obj:`torch.autograd.grad`.
    baseline_fn: callable, optional
        The baseline for which the model output is zero, supplied as a function of the input. Defaults to
        `torch.zeros_like`.
    n_iter: int, optional
        The number of iterations used to estimate the integral, defaults to 20.

    References
    ----------
    .. [2] M. Sundararajan, A. Taly, and Q. Yan, “Axiomatic attribution for deep networks,” in Proceedings of the 34th
       International Conference on Machine Learning, ICML 2017, Sydney, NSW, Australia, 6-11 August 2017, ser.
       Proceedings of Machine Learning Research, D. Precup and Y. W. Teh, Eds., vol. 70. PMLR, 2017, pp. 3319–3328.

    '''
    def __init__(
        self,
        model,
        composite=None,
        attr_output=None,
        create_graph=False,
        retain_graph=None,
        baseline_fn=None,
        n_iter=20
    ):
        super().__init__(
            model=model,
            composite=composite,
            attr_output=attr_output,
            create_graph=create_graph,
            retain_graph=retain_graph
        )
        if baseline_fn is None:
            baseline_fn = torch.zeros_like
        self.baseline_fn = baseline_fn
        self.n_iter = n_iter

    def forward(self, input, attr_output_fn):
        '''Compute the Integrated Gradients of the model wrt. input, by using `attr_output_fn` as the function of the
        model output to provide the vector for the vector jacobian product used to compute the gradient.
        This function will not register the composite, and is wrapped in the `__call__` of `Attributor`.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution.
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution function of the model's output.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model given `input`.
        attribution: :py:obj:`torch.Tensor`
            Attribution of the model wrt. to `input`, with the same shape as `input`.
        '''
        baseline = self.baseline_fn(input)

        result = torch.zeros_like(input)
        for alpha in torch.linspace(1. / self.n_iter, 1., self.n_iter):
            path_step = baseline + alpha * (input - baseline)
            output, gradient = self.grad(path_step, attr_output_fn)
            result += gradient / self.n_iter

        result *= (input - baseline)
        # in the last step, path_step is equal to input, thus `output` is the original output
        return output, result


class Occlusion(Attributor):
    '''This implements attribution by occlusion. Supplying a composite will have no effect on the result, as the
    gradient is not used.

    Parameters
    ----------
    model: :py:obj:`torch.nn.Module`
        The model for which the attribution will be computed. If `composite` is provided, this will also be the model
        to which the composite will be registered within `with` statements, or when calling the `Attributor` instance.
    composite: :py:obj:`zennit.core.Composite`, optional
        The optional composite to, if provided, be registered to the model within `with` statements or when calling the
        `Attributor` instance. Note that for Occlusion, this has no effect on the result.
    attr_output: :py:obj:`torch.Tensor` or callable, optional
        The default output attribution to be used when calling the `Attributor` instance, which is either a Tensor
        compatible with any input used, or a function of the model's output. If None (default), the value will be the
        identity function.
    occlusion_fn: callable, optional
        The occluded function, called with `occlusion_fn(input, mask)`, where `mask` is 1 inside the sliding window,
        and 0 outside. Either values inside or outside the sliding window may be occluded for different effects. By
        default, all values except inside the sliding window will be occluded.
    window: int or tuple of ints, optional
        The size of the sliding window to occlude over the input for each dimension. Defaults to 8. If a single integer
        is provided, the sliding window will slide with the same size over all dimensions except the first, which is
        assumed as the batch-dimension. If a tuple is provided, the window will only slide over the n-last dimensions,
        where n is the length of the tuple, e.g., if the data has shape `(3, 32, 32)` and `window=(8, 8)`, the
        resulting mask will have a block of shape `(3, 8, 8)` set to True. `window` must have the same length as
        `stride`.
    stride: int or tuple of ints, optional
        The strides used for the sliding window to occlude over the input for each dimension. Defaults to 8. If a
        single integer is provided, the strides will be the same size for all dimensions. If a tuple is provided,
        the window will only stride over the n-last dimensions, where n is the length of the tuple.
        `stride` must have the same length as `window`.

    '''
    def __init__(self, model, composite=None, attr_output=None, occlusion_fn=None, window=8, stride=8):
        def typecheck(obj):
            return isinstance(obj, int) or isinstance(obj, tuple) and all(isinstance(elem, int) for elem in obj)

        super().__init__(model=model, composite=composite, attr_output=attr_output)

        if not typecheck(window):
            raise TypeError('Occlusion window must either be an int, or a tuple of ints.')
        if not typecheck(stride):
            raise TypeError('Occlusion window must either be an int, or a tuple of ints.')

        if occlusion_fn is None:
            occlusion_fn = partial(occlude_independent, fill_fn=torch.zeros_like, invert=False)
        self.occlusion_fn = occlusion_fn
        self.window = window
        self.stride = stride

    def _resolve_window_stride(self, input):
        window = self.window
        stride = self.stride
        if isinstance(window, int):
            window = tuple(min(window, size) for size in input.shape[1:])
        if isinstance(stride, int):
            stride = tuple(min(stride, size) for size in input.shape[1:])

        if len(window) < input.ndim:
            window = tuple(input.shape)[:input.ndim - len(window)] + window
        if len(stride) < len(window):
            stride = window[:len(window) - len(stride)] + stride

        return window, stride

    def forward(self, input, attr_output_fn):
        '''Compute the occlusion analysis of the model wrt. input, by using `attr_output_fn` as function of the output,
        to return a weighting, which when multiplied again with the output, results in the classification score.
        This function will not register the composite, and is wrapped in the `__call__` of `Attributor`.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Input for the model, and wrt. compute the attribution.
        attr_output: :py:obj:`torch.Tensor` or callable, optional
            The output attribution function of the model's output.

        Returns
        -------
        output: :py:obj:`torch.Tensor`
            Output of the model given `input`.
        attribution: :py:obj:`torch.Tensor`
            Attribution of the model wrt. to `input`, with the same shape as `input`.
        '''
        window, stride = self._resolve_window_stride(input)

        root_mask = torch.zeros_like(input, dtype=bool)
        root_mask[tuple(slice(0, elem) for elem in window)] = True

        result = torch.zeros_like(input)
        for offset in product(*(range(0, size, local_stride) for size, local_stride in zip(input.shape, stride))):
            mask = root_mask.roll(offset, tuple(range(input.ndim)))
            occluded_input = self.occlusion_fn(input, mask)
            with torch.no_grad():
                output = self.model(occluded_input)
            score = attr_output_fn(output).sum(tuple(range(1, output.ndim)))
            result += mask * score[(slice(None),) + (None,) * (input.ndim - 1)]

        output = self.model(input)
        return output, result



'''Tests for torchvision-model-specific canonizers.'''
import pytest
import torch
from torchvision.models import vgg11_bn, resnet18, resnet50
from torchvision.models.resnet import BasicBlock as ResNetBasicBlock, Bottleneck as ResNetBottleneck
from helpers import assert_identity_hook, randomize_bnorm, nograd

from zennit.core import Composite, RemovableHandleList
from zennit.torchvision import VGGCanonizer, ResNetCanonizer
from zennit.types import BatchNorm


def test_vgg_canonizer(batchsize):
    '''Test whether VGGCanonizer merges BatchNorm modules correctly and keeps the output unchanged.'''
    model = randomize_bnorm(nograd(vgg11_bn().eval().to(torch.float64)))
    data = torch.randn((batchsize, 3, 224, 224), dtype=torch.float64)
    output_before = model(data)

    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(True, 'BatchNorm was not merged!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )

    canonizer = VGGCanonizer()
    composite = Composite(canonizers=[canonizer])

    try:
        composite.register(model)
        output_canonizer = model(data)
    finally:
        composite.remove()
        handles.remove()

    # this assumes the batch-norm is not initialized as the identity
    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(False, 'BatchNorm was not restored!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )
    try:
        output_after = model(data)
    finally:
        handles.remove()

    assert torch.allclose(output_canonizer, output_before, rtol=1e-5), 'Canonizer changed output after register!'
    assert torch.allclose(output_before, output_after, rtol=1e-5), 'Canonizer changed output after remove!'


@pytest.mark.parametrize('model_fn,block_type', [
    (resnet18, ResNetBasicBlock),
    (resnet50, ResNetBottleneck),
])
def test_resnet_canonizer(batchsize, model_fn, block_type):
    '''Test whether ResNetCanonizer overwrites and restores the Bottleneck/BasicBlock forward, merges BatchNorm modules
    correctly and keeps the output unchanged.
    '''
    model = randomize_bnorm(nograd(model_fn().eval().to(torch.float64)))
    data = torch.randn((batchsize, 3, 224, 224), dtype=torch.float64)
    blocks = [module for module in model.modules() if isinstance(module, block_type)]

    assert blocks, 'Model has no blocks!'
    assert all(
        block.forward == block_type.forward.__get__(block) for block in blocks
    ), 'Model has its forward already overwritten!'

    output_before = model(data)

    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(True, 'BatchNorm was not merged!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )

    canonizer = ResNetCanonizer()
    composite = Composite(canonizers=[canonizer])

    try:
        composite.register(model)
        assert not any(
            block.forward == block_type.forward.__get__(block) for block in blocks
        ), 'Model forward was not overwritten!'
        output_canonizer = model(data)
    finally:
        composite.remove()
        handles.remove()

    # this assumes the batch-norm is not initialized as the identity
    handles = RemovableHandleList(
        module.register_forward_hook(assert_identity_hook(False, 'BatchNorm was not restored!'))
        for module in model.modules() if isinstance(module, BatchNorm)
    )
    try:
        output_after = model(data)
    finally:
        handles.remove()

    assert all(
        block.forward == block_type.forward.__get__(block) for block in blocks
    ), 'Model forward was not restored!'
    assert torch.allclose(output_canonizer, output_before, rtol=1e-5), 'Canonizer changed output after register!'
    assert torch.allclose(output_before, output_after, rtol=1e-5), 'Canonizer changed output after remove!'



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/image.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Functionality to convert arrays to images'''
import numpy as np
from PIL import Image

from .cmap import LazyColorMapCache, ColorMap


# CMAPS contains all built-in color maps
CMAPS = LazyColorMapCache({
    # black to white
    'gray': '000,fff',
    # white to red
    'wred': 'fff,f00',
    # white to blue
    'wblue': 'fff,00f',
    # black to red to yellow to white
    'hot': '000,f00,ff0,fff',
    # black to blue to cyan
    'cold': '000,00f,0ff',
    # combination of cold (reversed) and hot, centered around black
    'coldnhot': '0ff,00f,80:000,f00,ff0,fff',
    # combination of wblue (reversed) and wred, centered around white
    'bwr': '00f,80:fff,f00',
    # blue to white to red as in the french flag
    'france': '0055a4,80:fff,ef4135',
    # blue to white to red with brightness 0xd0
    'seismic': '0000d0,80:d0d0d0,d00000',
    # cyan to white to magenta with brightness 0xd0
    'coolio': '00d0d0,80:d0d0d0,d000d0',
    # green to white to magenta with brightness 0xd0
    'coleus': '00d000,80:d0d0d0,d000d0',
})


def get_cmap(cmap):
    '''Convenience function to lookup built-in color maps, or create color maps from a source code.

    Parameters
    ----------
    cmap : str or ColorMap
        String to specify a built-in color map, code used to create a new color map, or a ColorMap instance.

    Returns
    -------
    ColorMap
        The built-in color map with key `cmap` in CMAPS, a new color map created from the code `cmap`, or `cmap` if it
        already was a ColorMap.
    '''
    if isinstance(cmap, ColorMap):
        return cmap
    if cmap in CMAPS:
        return CMAPS[cmap]
    return ColorMap(cmap)


def palette(cmap='bwr', level=1.0):
    '''Convenience function to create palettes from built-in colormaps, or from a source code if necessary.

    Parameters
    ----------
    cmap: str or ColorMap
        String to specify a built-in color map, code used to create a new color map, or a ColorMap instance, which will
        be used to create a palette.
    level: float
        The level of the color map palette. 1.0 is default. Values below zero reduce the color range, with only a
        single color used at value 0.0. Values above 1.0 clip the value earlier towards the maximum, with an
        increasingly steep transition at the center of the color map range.

    Returns
    -------
    obj:`numpy.ndarray`
        The palette described by an unsigned 8-bit numpy array with 256 entries.
    '''
    colormap = get_cmap(cmap)
    return colormap.palette(level=level)


def imgify(obj, vmin=None, vmax=None, cmap='bwr', level=1.0, symmetric=False, grid=False, gridfill=None):
    '''Convert an array with 1 or 3 channels to a PIL image.
    The color dimension can be either the first or the last dimension.

    Parameters
    ----------
    obj: object
        Anything that can be converted to a numpy array with 2 dimensions greyscale, or 3 dimensions with 1 or 3 values
        in the first or last dimension (color).
    vmin: float or obj:`numpy.ndarray`
        Manual minimum value of the array. Overrides the used norm's minimum value.
    vmax: float or obj:`numpy.ndarray`
        Manual maximum value of the array. Overrides the used norm's maximum value.
    cmap: str or ColorMap
        String to specify a built-in color map, code used to create a new color map, or a ColorMap instance, which will
        be used to create a palette. The color map will only be applied for arrays with only a single color channel.
        The color will be specified as a palette in the PIL Image.
    level: float
        The level of the color map. 1.0 is default. Values below 1.0 reduce the color range, with only a single color
        used at value 0.0. Values above 1.0 clip the value earlier towards the maximum, with an increasingly steep
        transition at the center of the pixel value distribution.
    symmetric : bool, optional
        Specifies whether the norm should be symmetric (True) or unaligned (False, default). If True, normalize with
        both minimum and maximum by the absolute maximum, which will cause 0. in the input to correspond to 0.5 in the
        result. Setting ``symmetric=False`` (default) will result in the minimum value to be directly mapped to 0 and
        the maximum value to be directly mapped to 1. ``vmin`` and ``vmax`` may be used to manually override the
        minimum and maximum value respectively.
    grid : bool or tuple of ints of size 2
        If true, assumes the first dimension to be the batch dimension. If True, creates a square grid of images in the
        batch dimension after normalizing each sample. If tuple of ints of size 2, creates the grid in the shape of
        ``(height, width)``. If False (default), does not assume a batch dimension.
    gridfill: :py:obj:`np.uint8`
        A value to fill empty grid members. Default is the mean pixel value. No effect when ``grid=False``.

    Returns
    -------
    image: obj:`PIL.Image`
        The array visualized as a Pillow Image.
    '''
    array = np.array(obj)

    if grid:
        if isinstance(grid, (list, tuple)) and len(grid) != 2:
            raise TypeError('Grid shape needs to be of size 2!')

        if array.ndim not in (3, 4):
            raise TypeError('Grid input has to have either 3 or 4 axes!')

        if (array.ndim == 4) and (array.shape[3] not in (1, 3)):
            if array.shape[1] in (1, 3):
                array = array.transpose(0, 2, 3, 1)
            else:
                raise TypeError(
                    'After batch, last (or first) axis of input are color channels, '
                    'which have to either be 1, 3 or be omitted entirely!'
                )
    else:
        if array.ndim not in (2, 3):
            raise TypeError('Input has to have either 2 or 3 axes!')

        if (array.ndim == 3) and (array.shape[2] not in (1, 3)):
            if array.shape[0] in (1, 3):
                array = array.transpose(1, 2, 0)
            else:
                raise TypeError(
                    'Last (or first) axis of input are color channels, '
                    'which have to either be 1, 3 or be omitted entirely!'
                )

    # renormalize data if necessary
    if array.dtype != np.uint8:
        if grid:
            dims = tuple(range(1, array.ndim))
        else:
            dims = tuple(range(array.ndim))

        if None in (vmin, vmax):
            vmin_, vmax_ = interval_norm_bounds(array, symmetric=symmetric, dim=dims)

        if vmin is None:
            vmin = vmin_
        else:
            vmin = np.array(vmin)

        if vmax is None:
            vmax = vmax_
        else:
            vmax = np.array(vmax)

        array = (array - vmin) / (vmax - vmin)
        array = (array * 255).clip(0, 255).astype(np.uint8)

    if grid:
        shape = None if isinstance(grid, bool) else grid
        # gridify adds the missing axis
        array = gridify(array, shape=shape, fill_value=gridfill)
    else:
        # add missing axis if omitted
        if array.ndim == 2:
            array = array[:, :, None]

    # apply palette if single channel
    if array.shape[2] == 1:
        image = Image.fromarray(array[..., 0], mode='P')
        image.putpalette(palette(cmap, level))
    else:
        image = Image.fromarray(array, mode='RGB')

    return image


def gridify(obj, shape=None, fill_value=None):
    '''Align multiple arrays, described as an additional 0-th dimension, into a grid with the 0-th dimension removed.

    Parameters
    ----------
    obj: object
        An object that can be converted to a numpy array, with 3 (greyscale) or 4 (rgb) axes.
        The color channel's position is automatically detected, and moved to the back of the shape.
    shape: tuple of size 2, optional
        Height and width of the produced grid. If None (default), create a square grid.
    fill_value: float or obj:`numpy.ndarray`
        A value to fill empty grid members. Default is the mean pixel value.

    Returns
    -------
    obj:`numpy.ndarray`
        An array with the 0-th dimension absorbed into the height and width dimensions (then 0 and 1).
        The color dimension will be the last dimension, even if it was the first dimension before.
    '''
    array = np.array(obj)
    if array.ndim not in (3, 4):
        raise TypeError('For creating an image grid, the array has to have either 3 (greyscale) or 4 (rgb) axes!')

    # add missing axis if omitted
    if array.ndim == 3:
        array = array[..., None]

    if array.shape[3] not in (1, 3):
        if array.shape[1] in (1, 3):
            array = array.transpose(0, 2, 3, 1)
        else:
            raise TypeError(
                'Last (or first) axis of input are color channels, '
                'which have to either be 1, 3 or be omitted entirely!'
            )

    num, height, width, channels = array.shape

    if shape is None:
        grid_width = int(num ** 0.5)
        grid_height = (num + grid_width - 1) // grid_width
    else:
        grid_height, grid_width = shape

    if fill_value is None:
        fill_value = array.min((0, 1, 2), keepdims=True)
    else:
        fill_value = np.array(fill_value).astype(array.dtype)

    dim = min(num, grid_height * grid_width)
    result = np.zeros((grid_height * grid_width, height, width, channels), dtype=array.dtype) + fill_value
    result[:dim] = array[:dim]
    result = (
        result
        .reshape(grid_height, grid_width, height, width, channels)
        .transpose(0, 2, 1, 3, 4)
        .reshape(grid_height * height, grid_width * width, channels)
    )

    return result


def imsave(
    fp,
    obj,
    vmin=None,
    vmax=None,
    cmap='bwr',
    level=1.0,
    grid=False,
    format=None,
    writer_params=None,
    symmetric=False,
    gridfill=None,
):
    '''Convert an array to an image and save it using file `fp`.
    Internally, `imgify` is called to create a PIL Image, which is then saved using PIL.

    Parameters
    ----------
    fp: str, obj:`pathlib.Path` or file
        Save target for PIL Image.
    obj: object
        Anything that can be converted to a numpy array with 2 dimensions greyscale, or 3 dimensions with 1 or 3 values
        in the first or last dimension (color).
    vmin: float or obj:`numpy.ndarray`
        Manual minimum value of the array. Overrides the used norm's minimum value.
    vmax: float or obj:`numpy.ndarray`
        Manual maximum value of the array. Overrides the used norm's maximum value.
    cmap: str or ColorMap
        String to specify a built-in color map, code used to create a new color map, or a ColorMap instance, which will
        be used to create a palette. The color map will only be applied for arrays with only a single color channel.
        The color will be specified as a palette in the PIL Image.
    level: float
        The level of the color map. 1.0 is default. Values below 1.0 reduce the color range, with only a single color
        used at value 0.0. Values above 1.0 clip the value earlier towards the maximum, with an increasingly steep
        transition at the center of the pixel value distribution.
    grid : bool or tuple of ints of size 2
        If true, assumes the first dimension to be the batch dimension. If True, creates a square grid of images in the
        batch dimension after normalizing each sample. If tuple of ints of size 2, creates the grid in the shape of
        ``(height, width)``. If False (default), does not assume a batch dimension.
    format: str
        Optional format override for PIL Image.save.
    writer_params: dict
        Extra params to the image writer in PIL.
    symmetric : bool, optional
        Specifies whether the norm should be symmetric (True) or unaligned (False, default). If True, normalize with
        both minimum and maximum by the absolute maximum, which will cause 0. in the input to correspond to 0.5 in the
        result. Setting ``symmetric=False`` (default) will result in the minimum value to be directly mapped to 0 and
        the maximum value to be directly mapped to 1. ``vmin`` and ``vmax`` may be used to manually override the
        minimum and maximum value respectively.
    gridfill: :py:obj:`np.uint8`
        A value to fill empty grid members. Default is the mean pixel value. No effect when ``grid=False``.
    '''
    if writer_params is None:
        writer_params = {}
    image = imgify(obj, vmin=vmin, vmax=vmax, cmap=cmap, level=level, symmetric=symmetric, grid=grid, gridfill=gridfill)
    image.save(fp, format=format, **writer_params)


def interval_norm_bounds(input, symmetric=False, dim=None):
    '''Return the boundaries to normalize the data interval batch-wise between 0. and 1. given the specified strategy.

    Parameters
    ----------
    input : :py:class:`numpy.ndarray`
        Array for which to return the boundaries.
    symmetric : bool, optional
        Specifies whether the norm should be symmetric (True) or unaligned (False, default).
        If True, normalize with both minimum and maximum by the absolute maximum, which will cause 0. in the
        input to correspond to 0.5 in the result. Setting ``symmetric=False`` (default) will result in the minimum
        value to be directly mapped to 0 and the maximum value to be directly mapped to 1.
    dim : tuple of ints, optional
        Set the channel dimensions over which the boundaries are computed (default is ``tuple(range(1, input.ndim))``.

    Returns
    -------
    :py:class:`numpy.ndarray`
        The normalized array ``input`` along ``dim`` to lie in the interval [0, 1].

    '''
    if dim is None:
        dim = tuple(range(1, input.ndim))

    if symmetric:
        # 0-aligned symmetric input, negative and positive can be compared, the original 0. becomes 0.5
        vmax = np.abs(input).max(dim, keepdims=True)
        vmin = -vmax
    else:
        # do not align, the original minimum value becomes 0., the original maximum becomes 1.
        vmax = input.max(dim, keepdims=True)
        vmin = input.min(dim, keepdims=True)
    return vmin, vmax



'''A quick example to generate heatmaps for vgg16.'''
import os
from functools import partial

import click
import torch
import numpy as np
from torch.utils.data import DataLoader, Subset
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor
from torchvision.datasets import ImageFolder
from torchvision.models import vgg16, vgg16_bn, resnet50

from zennit.attribution import Gradient, SmoothGrad, IntegratedGradients, Occlusion
from zennit.composites import COMPOSITES
from zennit.image import imsave, CMAPS
from zennit.torchvision import VGGCanonizer, ResNetCanonizer


MODELS = {
    'vgg16': (vgg16, VGGCanonizer),
    'vgg16_bn': (vgg16_bn, VGGCanonizer),
    'resnet50': (resnet50, ResNetCanonizer),
}

ATTRIBUTORS = {
    'gradient': Gradient,
    'smoothgrad': SmoothGrad,
    'integrads': IntegratedGradients,
    'occlusion': Occlusion,
    'inputxgrad': IntegratedGradients,
}


class BatchNormalize:
    def __init__(self, mean, std, device=None):
        self.mean = torch.tensor(mean, device=device)[None, :, None, None]
        self.std = torch.tensor(std, device=device)[None, :, None, None]

    def __call__(self, tensor):
        return (tensor - self.mean) / self.std


class AllowEmptyClassImageFolder(ImageFolder):
    '''Subclass of ImageFolder, which only finds non-empty classes, but with their correct indices given other empty
    classes. This counter-acts the changes in torchvision 0.10.0, in which DatasetFolder does not allow empty classes
    anymore by default. Versions before 0.10.0 do not expose `find_classes`, and thus this change does not change the
    functionality of `ImageFolder` in earlier versions.
    '''
    def find_classes(self, directory):
        with os.scandir(directory) as scanit:
            class_info = sorted((entry.name, len(list(os.scandir(entry.path)))) for entry in scanit if entry.is_dir())
        class_to_idx = {class_name: index for index, (class_name, n_members) in enumerate(class_info) if n_members}
        if not class_to_idx:
            raise FileNotFoundError(f'No non-empty classes found in \'{directory}\'.')
        return list(class_to_idx), class_to_idx


@click.command()
@click.argument('dataset-root', type=click.Path(file_okay=False))
@click.argument('relevance_format', type=click.Path(dir_okay=False, writable=True))
@click.option('--attributor', 'attributor_name', type=click.Choice(list(ATTRIBUTORS)), default='gradient')
@click.option('--composite', 'composite_name', type=click.Choice(list(COMPOSITES)))
@click.option('--model', 'model_name', type=click.Choice(list(MODELS)), default='vgg16_bn')
@click.option('--parameters', type=click.Path(dir_okay=False))
@click.option(
    '--inputs',
    'input_format',
    type=click.Path(dir_okay=False, writable=True),
    help='Input image format string.  {sample} is replaced with the sample index.'
)
@click.option('--batch-size', type=int, default=16)
@click.option('--max-samples', type=int)
@click.option('--n-outputs', type=int, default=1000)
@click.option('--cpu/--gpu', default=True)
@click.option('--shuffle/--no-shuffle', default=False)
@click.option('--with-bias/--no-bias', default=True)
@click.option('--relevance-norm', type=click.Choice(['symmetric', 'absolute', 'unaligned']), default='symmetric')
@click.option('--cmap', type=click.Choice(list(CMAPS)), default='coldnhot')
@click.option('--level', type=float, default=1.0)
@click.option('--seed', type=int, default=0xDEADBEEF)
def main(
    dataset_root,
    relevance_format,
    attributor_name,
    composite_name,
    model_name,
    parameters,
    input_format,
    batch_size,
    max_samples,
    n_outputs,
    cpu,
    shuffle,
    with_bias,
    cmap,
    level,
    relevance_norm,
    seed
):
    '''Generate heatmaps of an image folder at DATASET_ROOT to files RELEVANCE_FORMAT.
    RELEVANCE_FORMAT is a format string, for which {sample} is replaced with the sample index.
    '''
    # set a manual seed for the RNG
    torch.manual_seed(seed)

    # use the gpu if requested and available, else use the cpu
    device = torch.device('cuda:0' if torch.cuda.is_available() and not cpu else 'cpu')

    # mean and std of ILSVRC2012 as computed for the torchvision models
    norm_fn = BatchNormalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225), device=device)

    # transforms as used for torchvision model evaluation
    transform = Compose([
        Resize(256),
        CenterCrop(224),
        ToTensor(),
    ])

    # the dataset is a folder containing folders with samples, where each folder corresponds to one label
    dataset = AllowEmptyClassImageFolder(dataset_root, transform=transform)

    # limit the number of output samples, if requested, by creating a subset
    if max_samples is not None:
        if shuffle:
            indices = sorted(np.random.choice(len(dataset), min(len(dataset), max_samples), replace=False))
        else:
            indices = range(min(len(dataset), max_samples))
        dataset = Subset(dataset, indices)

    loader = DataLoader(dataset, shuffle=shuffle, batch_size=batch_size)

    model = MODELS[model_name][0]()

    # load model parameters if requested; the parameter file may need to be downloaded separately
    if parameters is not None:
        state_dict = torch.load(parameters)
        model.load_state_dict(state_dict)
    model.to(device)
    model.eval()

    # disable requires_grad for all parameters, we do not need their modified gradients
    for param in model.parameters():
        param.requires_grad = False

    # convenience identity matrix to produce one-hot encodings
    eye = torch.eye(n_outputs, device=device)

    # function to compute output relevance given the function output and a target
    def attr_output_fn(output, target):
        # output times one-hot encoding of the target labels of size (len(target), 1000)
        return output * eye[target]

    # create a composite if composite_name was set, otherwise we do not use a composite
    composite = None
    if composite_name is not None:
        composite_kwargs = {}
        if composite_name == 'epsilon_gamma_box':
            # the maximal input shape, needed for the ZBox rule
            shape = (batch_size, 3, 224, 224)

            # the highest and lowest pixel values for the ZBox rule
            composite_kwargs['low'] = norm_fn(torch.zeros(*shape, device=device))
            composite_kwargs['high'] = norm_fn(torch.ones(*shape, device=device))

        # provide the name 'bias' in zero_params if no bias should be used to compute the relevance
        if not with_bias and composite_name in [
            'epsilon_gamma_box',
            'epsilon_plus',
            'epsilon_alpha2_beta1',
            'epsilon_plus_flat',
            'epsilon_alpha2_beta1_flat',
            'excitation_backprop',
        ]:
            composite_kwargs['zero_params'] = ['bias']

        # use torchvision specific canonizers, as supplied in the MODELS dict
        composite_kwargs['canonizers'] = [MODELS[model_name][1]()]

        # create a composite specified by a name; the COMPOSITES dict includes all preset composites provided by zennit.
        composite = COMPOSITES[composite_name](**composite_kwargs)

    # specify some attributor-specific arguments
    attributor_kwargs = {
        'smoothgrad': {'noise_level': 0.1, 'n_iter': 20},
        'integrads': {'n_iter': 20},
        'inputxgrad': {'n_iter': 1},
        'occlusion': {'window': (56, 56), 'stride': (28, 28)},
    }.get(attributor_name, {})

    # create an attributor, given the ATTRIBUTORS dict given above. If composite is None, the gradient will not be
    # modified for the attribution
    attributor = ATTRIBUTORS[attributor_name](model, composite, **attributor_kwargs)

    # the current sample index for creating file names
    sample_index = 0

    # the accuracy
    accuracy = 0.

    # enter the attributor context outside the data loader loop, such that its canonizers and hooks do not need to be
    # registered and removed for each step. This registers the composite (and applies the canonizer) to the model
    # within the with-statement
    with attributor:
        for data, target in loader:
            # we use data without the normalization applied for visualization, and with the normalization applied as
            # the model input
            data_norm = norm_fn(data.to(device))

            # create output relevance function of output with fixed target
            output_relevance = partial(attr_output_fn, target=target)

            # this will compute the modified gradient of model, where the output relevance is chosen by the as the
            # model's output for the ground-truth label index
            output, relevance = attributor(data_norm, output_relevance)

            # sum over the color channel for visualization
            relevance = np.array(relevance.sum(1).detach().cpu())

            # normalize between 0. and 1. given the specified strategy
            if relevance_norm == 'symmetric':
                # 0-aligned symmetric relevance, negative and positive can be compared, the original 0. becomes 0.5
                amax = np.abs(relevance).max((1, 2), keepdims=True)
                relevance = (relevance + amax) / 2 / amax
            elif relevance_norm == 'absolute':
                # 0-aligned absolute relevance, only the amplitude of relevance matters, the original 0. becomes 0.
                relevance = np.abs(relevance)
                relevance /= relevance.max((1, 2), keepdims=True)
            elif relevance_norm == 'unaligned':
                # do not align, the original minimum value becomes 0., the original maximum becomes 1.
                rmin = relevance.min((1, 2), keepdims=True)
                rmax = relevance.max((1, 2), keepdims=True)
                relevance = (relevance - rmin) / (rmax - rmin)

            for n in range(len(data)):
                fname = relevance_format.format(sample=sample_index + n)
                # zennit.image.imsave will create an appropriate heatmap given a cmap specification
                imsave(fname, relevance[n], vmin=0., vmax=1., level=level, cmap=cmap)
                if input_format is not None:
                    fname = input_format.format(sample=sample_index + n)
                    # if there are 3 color channels, imsave will not create a heatmap, but instead save the image with
                    # its appropriate colors
                    imsave(fname, data[n])
            sample_index += len(data)

            # update the accuracy
            accuracy += (output.argmax(1) == target).sum().detach().cpu().item()

    accuracy /= len(dataset)
    print(f'Accuracy: {accuracy:.2f}')


if __name__ == '__main__':
    main()



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/types.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Type definitions for convenience.'''
import torch


class SubclassMeta(type):
    '''Meta class to bundle multiple subclasses.'''
    def __instancecheck__(cls, inst):
        """Implement isinstance(inst, cls) as subclasscheck."""
        return cls.__subclasscheck__(type(inst))

    def __subclasscheck__(cls, sub):
        """Implement issubclass(sub, cls) with by considering additional __subclass__ members."""
        candidates = cls.__dict__.get("__subclass__", tuple())
        return type.__subclasscheck__(cls, sub) or issubclass(sub, candidates)


class ConvolutionTranspose(metaclass=SubclassMeta):
    '''Abstract base class that describes transposed convolutional modules.'''
    __subclass__ = (
        torch.nn.modules.conv.ConvTranspose1d,
        torch.nn.modules.conv.ConvTranspose2d,
        torch.nn.modules.conv.ConvTranspose3d,
    )


class ConvolutionStandard(metaclass=SubclassMeta):
    '''Abstract base class that describes standard (forward) convolutional modules.'''
    __subclass__ = (
        torch.nn.modules.conv.Conv1d,
        torch.nn.modules.conv.Conv2d,
        torch.nn.modules.conv.Conv3d,
    )


class Convolution(metaclass=SubclassMeta):
    '''Abstract base class that describes all convolutional modules.'''
    __subclass__ = (
        ConvolutionStandard,
        ConvolutionTranspose,
    )


class Linear(metaclass=SubclassMeta):
    '''Abstract base class that describes linear modules.'''
    __subclass__ = (
        Convolution,
        torch.nn.modules.linear.Linear,
    )


class BatchNorm(metaclass=SubclassMeta):
    '''Abstract base class that describes batch normalization modules.'''
    __subclass__ = (
        torch.nn.modules.batchnorm.BatchNorm1d,
        torch.nn.modules.batchnorm.BatchNorm2d,
        torch.nn.modules.batchnorm.BatchNorm3d,
    )


class AvgPool(metaclass=SubclassMeta):
    '''Abstract base class that describes sum-pooling modules.'''
    __subclass__ = (
        torch.nn.modules.pooling.AvgPool1d,
        torch.nn.modules.pooling.AvgPool2d,
        torch.nn.modules.pooling.AvgPool3d,
        torch.nn.modules.pooling.AdaptiveAvgPool1d,
        torch.nn.modules.pooling.AdaptiveAvgPool2d,
        torch.nn.modules.pooling.AdaptiveAvgPool3d,
    )


class MaxPool(metaclass=SubclassMeta):
    '''Abstract base class that describes max-pooling modules.'''
    __subclass__ = (
        torch.nn.modules.pooling.MaxPool1d,
        torch.nn.modules.pooling.MaxPool2d,
        torch.nn.modules.pooling.MaxPool3d,
        torch.nn.modules.pooling.AdaptiveMaxPool1d,
        torch.nn.modules.pooling.AdaptiveMaxPool2d,
        torch.nn.modules.pooling.AdaptiveMaxPool3d,
    )


class Activation(metaclass=SubclassMeta):
    '''Abstract base class that describes activation modules.'''
    __subclass__ = (
        torch.nn.modules.activation.ELU,
        torch.nn.modules.activation.Hardshrink,
        torch.nn.modules.activation.Hardsigmoid,
        torch.nn.modules.activation.Hardtanh,
        torch.nn.modules.activation.Hardswish,
        torch.nn.modules.activation.LeakyReLU,
        torch.nn.modules.activation.LogSigmoid,
        torch.nn.modules.activation.MultiheadAttention,
        torch.nn.modules.activation.PReLU,
        torch.nn.modules.activation.ReLU,
        torch.nn.modules.activation.ReLU6,
        torch.nn.modules.activation.RReLU,
        torch.nn.modules.activation.SELU,
        torch.nn.modules.activation.CELU,
        torch.nn.modules.activation.GELU,
        torch.nn.modules.activation.Sigmoid,
        torch.nn.modules.activation.SiLU,
        torch.nn.modules.activation.Softplus,
        torch.nn.modules.activation.Softshrink,
        torch.nn.modules.activation.Softsign,
        torch.nn.modules.activation.Tanh,
        torch.nn.modules.activation.Tanhshrink,
        torch.nn.modules.activation.Threshold,
    )



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/torchvision.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Specialized Canonizers for models from torchvision.'''
import torch
from torchvision.models.resnet import Bottleneck as ResNetBottleneck, BasicBlock as ResNetBasicBlock

from .canonizers import SequentialMergeBatchNorm, AttributeCanonizer, CompositeCanonizer
from .layer import Sum


class VGGCanonizer(SequentialMergeBatchNorm):
    '''Canonizer for torchvision.models.vgg* type models. This is so far identical to a SequentialMergeBatchNorm'''


class ResNetBottleneckCanonizer(AttributeCanonizer):
    '''Canonizer specifically for Bottlenecks of torchvision.models.resnet* type models.'''
    def __init__(self):
        super().__init__(self._attribute_map)

    @classmethod
    def _attribute_map(cls, name, module):
        '''Create a forward function and a Sum module to overload as new attributes for module.

        Parameters
        ----------
        name : string
            Name by which the module is identified.
        module : obj:`torch.nn.Module`
            Instance of a module. If this is a Bottleneck layer, the appropriate attributes to overload are returned.

        Returns
        -------
        None or dict
            None if `module` is not an instance of Bottleneck, otherwise the appropriate attributes to overload onto
            the module instance.
        '''
        if isinstance(module, ResNetBottleneck):
            attributes = {
                'forward': cls.forward.__get__(module),
                'canonizer_sum': Sum(),
            }
            return attributes
        return None

    @staticmethod
    def forward(self, x):
        '''Modified Bottleneck forward for ResNet.'''
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = torch.stack([identity, out], dim=-1)
        out = self.canonizer_sum(out)

        out = self.relu(out)

        return out


class ResNetBasicBlockCanonizer(AttributeCanonizer):
    '''Canonizer specifically for BasicBlocks of torchvision.models.resnet* type models.'''
    def __init__(self):
        super().__init__(self._attribute_map)

    @classmethod
    def _attribute_map(cls, name, module):
        '''Create a forward function and a Sum module to overload as new attributes for module.

        Parameters
        ----------
        name : string
            Name by which the module is identified.
        module : obj:`torch.nn.Module`
            Instance of a module. If this is a BasicBlock layer, the appropriate attributes to overload are returned.

        Returns
        -------
        None or dict
            None if `module` is not an instance of BasicBlock, otherwise the appropriate attributes to overload onto
            the module instance.
        '''
        if isinstance(module, ResNetBasicBlock):
            attributes = {
                'forward': cls.forward.__get__(module),
                'canonizer_sum': Sum(),
            }
            return attributes
        return None

    @staticmethod
    def forward(self, x):
        '''Modified BasicBlock forward for ResNet.'''
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = torch.stack([identity, out], dim=-1)
        out = self.canonizer_sum(out)

        out = self.relu(out)

        return out


class ResNetCanonizer(CompositeCanonizer):
    '''Canonizer for torchvision.models.resnet* type models. This applies SequentialMergeBatchNorm, as well as
    add a Sum module to the Bottleneck modules and overload their forward method to use the Sum module instead of
    simply adding two tensors, such that forward and backward hooks may be applied.'''
    def __init__(self):
        super().__init__((
            SequentialMergeBatchNorm(),
            ResNetBottleneckCanonizer(),
            ResNetBasicBlockCanonizer(),
        ))



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/core.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Core functions and classes'''
import functools
import weakref
from contextlib import contextmanager

import torch


class Stabilizer:
    '''Class to create a stabilizer callable.

    Parameters
    ----------
    epsilon: float, optional
        Value by which to shift/clip elements of ``input``.
    clip: bool, optional
        If ``False`` (default), add ``epsilon`` multiplied by each entry's sign (+1 for 0). If ``True``, instead clip
        the absolute value of ``input`` and multiply it by each entry's original sign.
    norm_scale: bool, optional
        If ``False`` (default), ``epsilon`` is added to/used to clip ``input``. If ``True``, scale ``epsilon`` by the
        square root of the mean over the squared elements of the specified dimensions ``dim``.
    dim: tuple[int], optional
        If ``norm_scale`` is ``True``, specifies the dimension over which the scaled norm should be computed (all
        except dimension 0 by default).

    '''
    def __init__(self, epsilon=1e-6, clip=False, norm_scale=False, dim=None):
        self.epsilon = epsilon
        self.clip = clip
        self.norm_scale = norm_scale
        self.dim = dim

    def __call__(self, input):
        '''Stabilize input for safe division. This shifts zero-elements by ``+ epsilon``. For the sake of the
        *epsilon rule*, this also shifts positive values by ``+ epsilon`` and negative values by ``- epsilon``.

        Parameters
        ----------
        input: :py:obj:`torch.Tensor`
            Tensor to stabilize.

        Returns
        -------
        :py:obj:`torch.Tensor`
            Stabilized ``input``.
        '''
        return stabilize(input, self.epsilon, self.clip, self.norm_scale, self.dim)

    @classmethod
    def ensure(cls, value):
        '''Given a value, return a stabilizer. If ``value`` is a float, a Stabilizer with that epsilon ``value`` is
        returned. If ``value`` is callable, it will be used directly as a stabilizer. Otherwise a TypeError will be
        raised.

        Parameters
        ----------
        value: float, int, or callable
            The value used to produce a valid stabilizer function.

        Returns
        -------
        callable or Stabilizer
            A callable to be used as a stabilizer.

        Raises
        ------
        TypeError
            If no valid stabilizer could be produced from ``value``.
        '''
        if isinstance(value, (float, int)):
            return cls(epsilon=float(value))
        if callable(value):
            return value
        raise TypeError(f'Value {value} is not a valid stabilizer!')


def stabilize(input, epsilon=1e-6, clip=False, norm_scale=False, dim=None):
    '''Stabilize input for safe division.

    Parameters
    ----------
    input: :py:obj:`torch.Tensor`
        Tensor to stabilize.
    epsilon: float, optional
        Value by which to shift/clip elements of ``input``.
    clip: bool, optional
        If ``False`` (default), add ``epsilon`` multiplied by each entry's sign (+1 for 0). If ``True``, instead clip
        the absolute value of ``input`` and multiply it by each entry's original sign.
    norm_scale: bool, optional
        If ``False`` (default), ``epsilon`` is added to/used to clip ``input``. If ``True``, scale ``epsilon`` by the
        square root of the mean over the squared elements of the specified dimensions ``dim``.
    dim: tuple[int], optional
        If ``norm_scale`` is ``True``, specifies the dimension over which the scaled norm should be computed. Defaults
        to all except dimension 0.

    Returns
    -------
    :py:obj:`torch.Tensor`
        New Tensor copied from `input` with values shifted by epsilon.
    '''
    sign = ((input == 0.).to(input) + input.sign())
    if norm_scale:
        if dim is None:
            dim = tuple(range(1, input.ndim))
        epsilon = epsilon * ((input ** 2).mean(dim=dim, keepdim=True) ** .5)
    if clip:
        return sign * input.abs().clip(min=epsilon)
    return input + sign * epsilon


def expand(tensor, shape, cut_batch_dim=False):
    '''Expand a scalar value or tensor to a shape. In addition to torch.Tensor.expand, this will also accept
    non-torch.tensor objects, which will be used to create a new tensor. If ``tensor`` has fewer dimensions than
    ``shape``, singleton dimension will be appended to match the size of ``shape`` before expanding.

    Parameters
    ----------
    tensor : int, float or :py:obj:`torch.Tensor`
        Scalar or tensor to expand to the size of ``shape``.
    shape : tuple[int]
        Shape to which ``tensor`` will be expanded.
    cut_batch_dim : bool, optional
        If True, take only the first ``shape[0]`` entries along dimension 0 of the expanded ``tensor``, if it has more
        entries in dimension 0 than ``shape``. Default (False) is not to cut, which will instead cause a
        ``RuntimeError`` due to the size mismatch.

    Returns
    -------
    :py:obj:`torch.Tensor`
        A new tensor expanded from ``tensor`` with shape ``shape``.

    Raises
    ------
    RuntimeError
        If ``tensor`` could not be expanded to ``shape`` due to incompatible shapes.

    '''
    if not isinstance(tensor, torch.Tensor):
        # cast non-tensor scalar to 0-dim tensor
        tensor = torch.tensor(tensor)
    if tensor.ndim == 0:
        # expand scalar tensors
        return tensor.expand(shape)
    if tensor.ndim < len(shape) and all(left in (1, right) for left, right in zip(tensor.shape, shape)):
        # append singleton dimensions if tensor has fewer dimensions, and the existing ones match with shape
        tensor = tensor[(...,) + (None,) * (len(shape) - len(tensor.shape))]
    if tensor.ndim == len(shape):
        # if the dims match completely (lengths match and zipped match), expand normally
        if all(left in (1, right) for left, right in zip(tensor.shape, shape)):
            return tensor.expand(shape)
        # if `cut_batch_dim` and dims match except first, which is larger than shape, the the first dim and expand
        if (
            cut_batch_dim
            and all(left in (1, right) for left, right in zip(tensor.shape[1:], shape[1:]))
            and tensor.shape[0] > shape[0]
        ):
            return tensor[:shape[0]].expand(shape)
    raise RuntimeError(f'Invalid shape! Target: {tensor.shape}; Source: {shape}')


def zero_wrap(zero_params):
    '''Create a function wrapper factory (i.e. a decorator), which takes a single function argument ``(name, param) ->
    tensor`` such that the function is only called if name is not equal to zero_params, if zero_params is a string, or
    it is not in zero_params. Otherwise return `torch.zeros_like` of that tensor.

    Parameters
    ----------
    zero_params: str or list[str]
        String or list of strings compared to `name`.

    Returns
    -------
    function
        The function wrapper to be called on the function.
    '''
    def zero_params_wrapper(modifier):
        '''Wrap a function (name, param) -> tensor such that the function is only called if name is not equal to the
        closure list zero_params, if zero_params is a string, or it is not in zero_params. Otherwise return
        `torch.zeros_like` of that tensor.

        Parameters
        ----------
        func: function
            Function to wrap.

        Returns
        -------
        function
            The wrapped function.
        '''
        if not zero_params:
            return modifier

        @functools.wraps(modifier)
        def modifier_wrapper(input, name):
            '''Wrapped function (name, param) -> tensor, where the original function is only called if name is not
            equal to the closure list zero_params, if zero_params is a string, or it is not in zero_params. Otherwise
            return `torch.zeros_like` of that tensor.

            Parameters
            ----------
            input: :py:obj:`torch.Tensor`
                The input tensor modified by the original function.
            name: str
                The name associated with the input tensor (e.g. the parameter name).

            Returns
            -------
            :py:obj:`torch.Tensor`
                The modified tensor.
            '''
            if isinstance(zero_params, str) and name == zero_params or name in zero_params:
                return torch.zeros_like(input)
            return modifier(input, name)

        return modifier_wrapper
    return zero_params_wrapper


class ParamMod:
    '''Class to produce a context manager to temporarily modify parameter attributes (all by default) of a module.

    Parameters
    ----------
    modifier: function
        A function used to modify parameter attributes. If `param_keys` is empty, this is not used.
    param_keys: list[str], optional
        A list of parameter names that shall be modified. If `None` (default), all parameters are modified (which may
        be none). If `[]`, no parameters are modified and `modifier` is ignored.
    zero_params: list[str], optional
        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
    require_params: bool, optional
        Whether existence of `module`'s params is mandatory (True by default). If the attribute exists but is `None`,
        it is not considered missing, and the modifier is not applied.
    '''
    def __init__(self, modifier, param_keys=None, zero_params=None, require_params=True):
        self.modifier = modifier
        self.param_keys = param_keys
        self.zero_params = zero_params
        self.require_params = require_params

    @classmethod
    def ensure(cls, modifier):
        '''If ``modifier`` is an instance of ParamMod, return it as-is, if it is callable, create a new instance with
        ``modifier`` as the ParamMod's function, otherwise raise a TypeError.

        Parameters
        ----------
        modifier : :py:obj:`ParamMod` or callable
            The modifier which, if necessary, will be used to construct a ParamMod.

        Returns
        -------
        :py:obj:`ParamMod`
            Either ``modifier`` as is, or a :py:obj:`ParamMod` constructed using ``modifier``.

        Raises
        ------
        TypeError
            If ``modifier`` is neither an instance of :py:obj:`ParamMod`, nor callable.
        '''
        if isinstance(modifier, cls):
            return modifier
        if callable(modifier):
            return cls(modifier)
        raise TypeError(f'{modifier} is neither an instance of {cls}, nor callable!')

    @contextmanager
    def __call__(self, module):
        '''Context manager to temporarily modify parameter attributes (all by default) of a module.

        Parameters
        ----------
        module: :py:obj:`torch.nn.Module`
            Module of which to modify parameters. If `self.requires_params` is `True`, it must have all elements given
            in `self.param_keys` as attributes (attributes are allowed to be `None`, in which case they are ignored).

        Raises
        ------
        RuntimeError
            If `self.require_params` is `True` and `module` is missing an attribute listed in `self.param_keys`.

        Yields
        ------
        module: :py:obj:`torch.nn.Module`
            The `module` with appropriate parameters temporarily modified.
        '''
        try:
            stored_params = {}
            param_keys = self.param_keys
            zero_params = self.zero_params

            if param_keys is None:
                param_keys = [name for name, _ in module.named_parameters(recurse=False)]
            if zero_params is None:
                zero_params = []

            missing = [key for key in param_keys if not hasattr(module, key)]
            if self.require_params and missing:
                missing_str = '\', \''.join(missing)
                raise RuntimeError(f'Module {module} requires missing parameters: \'{missing_str}\'')

            modifier = zero_wrap(zero_params)(self.modifier)

            for key in param_keys:
                if key not in missing:
                    param = getattr(module, key)
                    if param is not None:
                        stored_params[key] = param
                        setattr(module, key, torch.nn.Parameter(modifier(param.data, key)))

            yield module
        finally:
            for key, value in stored_params.items():
                setattr(module, key, value)


def collect_leaves(module):
    '''Generator function to collect all leaf modules of a module.

    Parameters
    ----------
    module: :py:obj:`torch.nn.Module`
        A module for which the leaves will be collected.

    Yields
    ------
    leaf: :py:obj:`torch.nn.Module`
        Either a leaf of the module structure, or the module itself if it has no children.
    '''
    is_leaf = True

    children = module.children()
    for child in children:
        is_leaf = False
        for leaf in collect_leaves(child):
            yield leaf
    if is_leaf:  # pragma: no branch
        yield module


class Identity(torch.autograd.Function):
    '''Identity to add a grad_fn to a tensor, so a backward hook can be applied.'''
    @staticmethod
    def forward(ctx, *inputs):
        '''Forward identity.'''
        return inputs

    @staticmethod
    def backward(ctx, *grad_outputs):
        '''Backward identity.'''
        return grad_outputs


class Hook:
    '''Base class for hooks to be used to compute layer-wise attributions.'''
    def __init__(self):
        self.stored_tensors = {}
        self.active = True
        self.tensor_handles = RemovableHandleList()

    def pre_forward(self, module, input):
        '''Apply an Identity to the input before the module to register a backward hook.'''
        hook_ref = weakref.ref(self)

        @functools.wraps(self.backward)
        def wrapper(grad_input, grad_output):
            hook = hook_ref()
            if hook is not None and hook.active:
                return hook.backward(module, grad_input, hook.stored_tensors['grad_output'])
            return None

        if not isinstance(input, tuple):
            input = (input,)

        # only if gradient required
        if input[0].requires_grad:
            # add identity to ensure .grad_fn exists
            post_input = Identity.apply(*input)
            # register the input tensor gradient hook
            self.tensor_handles.append(
                post_input[0].grad_fn.register_hook(wrapper)
            )
            # work around to support in-place operations
            post_input = tuple(elem.clone() for elem in post_input)
        else:
            # no gradient required
            post_input = input
        return post_input[0] if len(post_input) == 1 else post_input

    def post_forward(self, module, input, output):
        '''Register a backward-hook to the resulting tensor right after the forward.'''
        hook_ref = weakref.ref(self)

        @functools.wraps(self.pre_backward)
        def wrapper(grad_input, grad_output):
            hook = hook_ref()
            if hook is not None and hook.active:
                return hook.pre_backward(module, grad_input, grad_output)
            return None

        if not isinstance(output, tuple):
            output = (output,)

        # only if gradient required
        if output[0].grad_fn is not None:
            # register the output tensor gradient hook
            self.tensor_handles.append(
                output[0].grad_fn.register_hook(wrapper)
            )
        return output[0] if len(output) == 1 else output

    def pre_backward(self, module, grad_input, grad_output):
        '''Store the grad_output for the backward hook'''
        self.stored_tensors['grad_output'] = grad_output

    def forward(self, module, input, output):
        '''Hook applied during forward-pass'''

    def backward(self, module, grad_input, grad_output):
        '''Hook applied during backward-pass'''

    def copy(self):
        '''Return a copy of this hook.
        This is used to describe hooks of different modules by a single hook instance.
        '''
        return self.__class__()

    def remove(self):
        '''When removing hooks, remove all references to stored tensors'''
        self.stored_tensors.clear()
        self.tensor_handles.remove()

    def register(self, module):
        '''Register this instance by registering all hooks to the supplied module.'''
        return RemovableHandleList([
            RemovableHandle(self),
            module.register_forward_pre_hook(self.pre_forward),
            module.register_forward_hook(self.post_forward),
            module.register_forward_hook(self.forward),
        ])


class BasicHook(Hook):
    '''A hook to compute the layer-wise attribution of the module it is attached to.
    A BasicHook instance may only be registered with a single module.

    Parameters
    ----------
    input_modifiers: list[callable], optional
        A list of functions ``(input: torch.Tensor) -> torch.Tensor`` to produce multiple inputs. Default is a single
        input which is the identity.
    param_modifiers: list[:py:obj:`~zennit.core.ParamMod` or callable], optional
        A list of ParamMod instances or functions ``(obj: torch.Tensor, name: str) -> torch.Tensor``, with parameter
        tensor ``obj``, registered in the root model as ``name``, to temporarily modify the parameters of the attached
        module for each input produced with `input_modifiers`. Default is unmodified parameters for each input. Use a
        :py:obj:`~zennit.core.ParamMod` instance to specify which parameters should be modified, whether they are
        required, and which should be set to zero.
    output_modifiers: list[callable], optional
        A list of functions ``(input: torch.Tensor) -> torch.Tensor`` to modify the module's output computed using the
        modified parameters before gradient computation for each input produced with `input_modifier`. Default is the
        identity for each output.
    gradient_mapper: callable, optional
        Function ``(out_grad: torch.Tensor, outputs: list[torch.Tensor]) -> list[torch.Tensor]`` to modify upper
        relevance. A list or tuple of the same size as ``outputs`` is expected to be returned. ``outputs`` has the same
        size as ``input_modifiers`` and ``param_modifiers``. Default is a stabilized normalization by each of the
        outputs, multiplied with the output gradient.
    reducer: callable
        Function ``(inputs: list[torch.Tensor], gradients: list[torch.Tensor]) -> torch.Tensor`` to reduce all the
        inputs and gradients produced through ``input_modifiers`` and ``param_modifiers``. ``inputs`` and ``gradients``
        have the same as ``input_modifiers`` and ``param_modifiers``. Default is the sum of the multiplications of each
        input and its corresponding gradient.
    '''
    def __init__(
        self,
        input_modifiers=None,
        param_modifiers=None,
        output_modifiers=None,
        gradient_mapper=None,
        reducer=None,
        stabilizer=1e-6,
    ):
        super().__init__()
        modifiers = {
            'in': input_modifiers,
            'param': param_modifiers,
            'out': output_modifiers,
        }
        supplied = {key for key, val in modifiers.items() if val is not None}
        num_mods = len(modifiers[next(iter(supplied))]) if supplied else 1
        modifiers.update({key: (self._default_modifier,) * num_mods for key in set(modifiers) - supplied})

        if gradient_mapper is None:
            gradient_mapper = self._default_gradient_mapper
        if reducer is None:
            reducer = self._default_reducer

        self.input_modifiers = modifiers['in']
        self.param_modifiers = modifiers['param']
        self.output_modifiers = modifiers['out']
        self.gradient_mapper = gradient_mapper
        self.reducer = reducer

    def forward(self, module, input, output):
        '''Forward hook to save module in-/outputs.'''
        self.stored_tensors['input'] = input

    def backward(self, module, grad_input, grad_output):
        '''Backward hook to compute LRP based on the class attributes.'''
        original_input = self.stored_tensors['input'][0].clone()
        inputs = []
        outputs = []
        for in_mod, param_mod, out_mod in zip(self.input_modifiers, self.param_modifiers, self.output_modifiers):
            input = in_mod(original_input).requires_grad_()
            with ParamMod.ensure(param_mod)(module) as modified, torch.autograd.enable_grad():
                output = modified.forward(input)
                output = out_mod(output)
            inputs.append(input)
            outputs.append(output)
        grad_outputs = self.gradient_mapper(grad_output[0], outputs)
        gradients = torch.autograd.grad(
            outputs,
            inputs,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad
        )
        relevance = self.reducer(inputs, gradients)
        return tuple(relevance if original.shape == relevance.shape else None for original in grad_input)

    def copy(self):
        '''Return a copy of this hook.
        This is used to describe hooks of different modules by a single hook instance.
        '''
        copy = BasicHook.__new__(type(self))
        BasicHook.__init__(
            copy,
            self.input_modifiers,
            self.param_modifiers,
            self.output_modifiers,
            self.gradient_mapper,
            self.reducer,
        )
        return copy

    @staticmethod
    def _default_modifier(obj, name=None):
        return obj

    @staticmethod
    def _default_gradient_mapper(out_grad, outputs):
        return tuple(out_grad / stabilize(output) for output in outputs)

    @staticmethod
    def _default_reducer(inputs, gradients):
        return sum(input * gradient for input, gradient in zip(inputs, gradients))


class RemovableHandle:
    '''Create weak reference to call .remove on some instance.'''
    def __init__(self, instance):
        self.instance_ref = weakref.ref(instance)

    def remove(self):
        '''Call remove on weakly reference instance if it still exists.'''
        instance = self.instance_ref()
        if instance is not None:
            instance.remove()


class RemovableHandleList(list):
    '''A list to hold handles, with the ability to call remove on all of its members.'''
    def remove(self):
        '''Call remove on all members, effectively removing handles from modules, or reverting canonizers.'''
        for handle in self:
            handle.remove()
        self.clear()


class CompositeContext:
    '''A context object to register a composite in a context and remove the associated hooks and canonizers afterwards.

    Parameters
    ----------
    module: :py:class:`torch.nn.Module`
        The module to which `composite` should be registered.
    composite: :py:class:`zennit.core.Composite`
        The composite which shall be registered to `module`.
    '''
    def __init__(self, module, composite):
        self.module = module
        self.composite = composite

    def __enter__(self):
        self.composite.register(self.module)
        return self.module

    def __exit__(self, exc_type, exc_value, traceback):
        self.composite.remove()
        return False


class Composite:
    '''A Composite to apply canonizers and register hooks to modules.
    One Composite instance may only be applied to a single module at a time.

    Parameters
    ----------
    module_map: callable, optional
        A function ``(ctx: dict, name: str, module: torch.nn.Module) -> Hook or None`` which maps a context, name and
        module to a matching :py:class:`~zennit.core.Hook`, or ``None`` if there is no matchin
        :py:class:`~zennit.core.Hook`.
    canonizers: list[:py:class:`zennit.canonizers.Canonizer`], optional
        List of canonizer instances to be applied before applying hooks.
    '''
    def __init__(self, module_map=None, canonizers=None):
        if module_map is None:
            module_map = self._empty_module_map
        if canonizers is None:
            canonizers = []

        self.module_map = module_map
        self.canonizers = canonizers

        self.handles = RemovableHandleList()
        self.hook_refs = weakref.WeakSet()

    def register(self, module):
        '''Apply all canonizers and register all hooks to a module (and its recursive children).
        Previous canonizers of this composite are reverted and all hooks registered by this composite are removed.
        The module or any of its children (recursively) may still have other hooks attached.

        Parameters
        ----------
        module: :py:class:`torch.nn.Module`
            Hooks and canonizers will be applied to this module recursively according to ``module_map`` and
            ``canonizers``.
        '''
        self.remove()

        for canonizer in self.canonizers:
            self.handles += canonizer.apply(module)

        ctx = {}
        for name, child in module.named_modules():
            template = self.module_map(ctx, name, child)
            if template is not None:
                hook = template.copy()
                self.hook_refs.add(hook)
                self.handles.append(hook.register(child))

    def remove(self):
        '''Remove all handles for hooks and canonizers.
        Hooks will simply be removed from their corresponding Modules.
        Canonizers will revert the state of the modules they changed.
        '''
        self.handles.remove()
        self.hook_refs.clear()

    def context(self, module):
        '''Return a CompositeContext object with this instance and the supplied module.

        Parameters
        ----------
        module: :py:class:`torch.nn.Module`
            Module for which to register this composite in the context.

        Returns
        -------
        :py:class:`zennit.core.CompositeContext`
            A context object which registers the composite to ``module`` on entering, and removes it on exiting.
        '''
        return CompositeContext(module, self)

    @contextmanager
    def inactive(self):
        '''Context manager to temporarily deactivate the gradient modification. This can be used to compute the
        gradient of the modified gradient.
        '''
        try:
            for hook in self.hook_refs:
                hook.active = False
            yield self
        finally:
            for hook in self.hook_refs:
                hook.active = True

    @staticmethod
    def _empty_module_map(ctx, name, module):
        '''Empty module_map, does not assign any rules.'''
        return None



# This file is part of Zennit
# Copyright (C) 2019-2021 Christopher J. Anders
#
# zennit/cmap.py
#
# Zennit is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# Zennit is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
# more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library. If not, see <https://www.gnu.org/licenses/>.
'''Create color maps from a color-map specification language'''
import re
from typing import NamedTuple

import numpy as np


class CMapToken(NamedTuple):
    '''Tokens used by the lexer of ColorMap.'''
    type: str
    value: str
    pos: int


class ColorNode(NamedTuple):
    '''Nodes produced by the parser of ColorMap.'''
    index: int
    value: np.ndarray


class ColorMap:
    '''Compile a color map from color-map specification language (cmsl) source code.

    The color-map specification language (cmsl) is used to specify linear color maps with comma-separated colors
    supplied as hexadecimal values for each color channel in RGB, with either 1 or 2 values per channel.  Optionally, a
    hexadecimal index with either one or two digits may be supplied in front of each color, followed by a colon, to
    indicate the index which should be the color. Values for the ColorMap in-between colors will be interpolated
    linearly. If no index is supplied, colors without indices will be spaced evenly between indices. If the first and
    last indices are supplied but not 0 (or 00) and f (or ff) respectively, they will be added as an additional node in
    the color map, with the same color as the colors with the lowest and highest index respectively.  If indices are
    provided, they must be in ascending order from left to right, with an arbitrary number of non-indexed colors. If
    the first and/or last color are not indexed, they are assumed to be 0 (or 00) and f (or ff) respectively.

    Parameters
    ----------
    source : str
        Source code to generate the color map.

    '''
    _rexp = re.compile(
        r'(?P<longcolor>[0-9a-fA-F]{6})|'
        r'(?P<shortcolor>[0-9a-fA-F]{3})|'
        r'(?P<index>[0-9a-fA-F]{1,2})|'
        r'(?P<adsep>:)|'
        r'(?P<sep>,)|'
        r'(?P<whitespace>\s+)|'
        r'(?P<error>.+)'
    )

    def __init__(self, source):
        self._source = None
        self.source = source

    @property
    def source(self) -> str:
        '''Source code property used to generate the color map. May be overwritten with a new string, which will be
        compiled to change the color map.
        '''
        return self._source

    @source.setter
    def source(self, value: str):
        '''Set source code property and re-compile the color map.'''
        try:
            tokens = self._lex(value)
            nodes = self._parse(tokens)
            self._indices, self._colors = self._make_palette(nodes)
        except RuntimeError as err:
            raise RuntimeError('Compilation of ColorMap failed!') from err

        self._source = value

    @staticmethod
    def _lex(string):
        '''Lexical scanning of cmsl using regular expressions.'''
        return [CMapToken(match.lastgroup, match.group(), match.start()) for match in ColorMap._rexp.finditer(string)]

    @staticmethod
    def _parse(tokens):
        '''Parse cmsl tokens into a list of color nodes.'''
        nodes = []
        log = []
        for token in tokens:
            if token.type == 'index' and not log:
                log.append(token)
            elif token.type == 'adsep' and len(log) == 1 and log[-1].type == 'index':
                log.append(token)
            elif token.type in ('shortcolor', 'longcolor'):
                if len(log) == 2 and log[-2].type == 'index':
                    indval = log[-2].value
                    if len(indval) == 1:
                        indval = indval * 2
                    index = int(indval, base=16)
                elif not log:
                    index = None
                else:
                    raise RuntimeError(f'Unexpected {token}')

                value_it = iter(token.value) if token.type == 'longcolor' else token.value
                value = [int(''.join(chars), base=16) for chars in zip(*[value_it] * 2)]
                nodes.append(ColorNode(index, np.array(value)))
                log.append(token)
            elif token.type == 'sep' and log and log[-1].type in ('longcolor', 'shortcolor'):
                log.clear()
            elif token.type != 'whitespace':
                raise RuntimeError(f'Unexpected {token}')

        if log and log[-1].type not in ('shortcolor', 'longcolor'):
            raise RuntimeError(f'Unexpected {log[-1]}')

        return nodes

    @staticmethod
    def _make_palette(nodes):
        '''Generate color map indices and colors from a list of color nodes.'''
        if len(nodes) < 2:
            raise RuntimeError("ColorMap needs at least 2 colors!")
        result = []
        log = []

        start = nodes.pop(0)
        result.append(ColorNode(0, start.value))
        if start.index is not None and start.index > 0:
            result.append(start)

        for n, node in enumerate(nodes):
            if node.index is None:
                if n < len(nodes) - 1:
                    log.append(node)
                    continue
                node = ColorNode(255, node.value)
            elif node.index < result[-1].index:
                raise RuntimeError('ColorMap indices not ordered! Provided indices are required in ascending order.')
            if log:
                result += [
                    ColorNode(
                        int(result[-1].index * (1. - alpha) + node.index * alpha),
                        lognode.value
                    ) for alpha, lognode in zip(np.linspace(0., 1., len(log) + 2)[1:-1], log)
                ]
                log.clear()
            result.append(node)

        result.append(ColorNode(256, result[-1].value))

        indices = np.array([node.index for node in result])
        colors = np.stack([node.value for node in result], axis=0)

        return indices, colors

    def __call__(self, x):
        '''Map scalar values in the range [0, 1] to RGB. This appends an axis with size 3 to `x`. Values are clipped to
        the range [0, 1], and the output will also lie in this range.

        Parameters
        ----------
        x : obj:`numpy.ndarray`
            Input array of arbitrary shape, which will be clipped to range [0, 1], and mapped to RGB using this
            ColorMap.

        Returns
        -------
        obj:`numpy.ndarray`
            The input array `x`, clipped to [0, 1] and mapped to RGB given this colormap, where the 3 color channels
            are appended as a new axis to the end.
        '''
        x = (x * 255).clip(0, 255)
        index = np.searchsorted(self._indices[:-1], x, side='right')
        alpha = ((x - self._indices[index - 1]) / (self._indices[index] - self._indices[index - 1]))[..., None]
        return (self._colors[index - 1] * (1 - alpha) + self._colors[index] * alpha) / 255.

    def palette(self, level=1.0):
        '''Create an 8-bit palette.

        Parameters
        ----------
        level: float
            The level of the color map. 1.0 is default. Values below zero reduce the color range, with only a single
            color used at value 0.0. Values above 1.0 clip the value earlier towards the maximum, with an increasingly
            steep transition at the center of the image.

        Returns
        -------
        obj:`numpy.ndarray`
            The palette described by an unsigned 8-bit numpy array with 256 entries.
        '''
        x = np.linspace(-1., 1., 256, dtype=np.float64) * level
        x = ((x + 1.) / 2.).clip(0., 1.)
        x = self(x)
        x = (x * 255.).round(12).clip(0., 255.).astype(np.uint8)
        return x


class LazyColorMapCache:
    '''Dict-like object to store sources for colormaps, and compile and cache them lazily.

    Parameters
    ----------
    sources : dict
        Dict containing a mapping from names to color map specification language source.
    '''
    def __init__(self, sources):
        self._sources = sources
        self._compiled = {}

    def __getitem__(self, name):
        if name not in self._sources:
            raise KeyError(f'No source for key {name}.')
        if name not in self._compiled:
            self._compiled[name] = ColorMap(self._sources[name])
        return self._compiled[name]

    def __setitem__(self, name, value):
        self._sources[name] = value
        if name in self._compiled:
            self._compiled[name].source = value

    def __delitem__(self, name):
        del self._sources[name]
        if name in self._compiled:
            del self._compiled[name]

    def __iter__(self):
        return iter(self._sources)

    def __len__(self):
        return len(self._sources)



'''Tests for core functionality in zennit.core.'''
from itertools import product, islice

import torch
import pytest
from helpers import nograd, prodict

from zennit.core import stabilize, expand, ParamMod, collect_leaves, Stabilizer
from zennit.core import Identity, Hook, BasicHook, RemovableHandle, RemovableHandleList, Composite


@pytest.mark.parametrize('kwargs,input,expected', [
    (
        {},
        [0., -0., 1., -1.],
        [1e-6, 1e-6, 1. + 1e-6, -1. - 1e-6]
    ), (
        {'epsilon': 1e-3, 'clip': False, 'norm_scale': False},
        [0., -0., 1., -1.],
        [1e-3, 1e-3, 1. + 1e-3, -1. - 1e-3]
    ), (
        {'epsilon': 1., 'clip': False, 'norm_scale': False},
        [0., -0., 1., -1.],
        [1., 1., 2., -2.]
    ), (
        {'epsilon': 1e-6, 'clip': True, 'norm_scale': False},
        [0., -0., 1., -1.],
        [1e-6, 1e-6, 1., -1.]
    ), (
        {'epsilon': 1e-3, 'clip': True, 'norm_scale': False},
        [0., -0., 1., -1.],
        [1e-3, 1e-3, 1., -1.]
    ), (
        {'epsilon': 1., 'clip': True, 'norm_scale': False},
        [0., -0., 1., -1.],
        [1., 1., 1., -1.]
    ), (
        {'epsilon': 1e-6, 'clip': False, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142e-6, 1.4142e-6, 2. + 1.4142e-6, -2. - 1.4142e-6]
    ), (
        {'epsilon': 1e-3, 'clip': False, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142e-3, 1.4142e-3, 2. + 1.4142e-3, -2. - 1.4142e-3]
    ), (
        {'epsilon': 1., 'clip': False, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142, 1.4142, 3.4142, -3.4142]
    ), (
        {'epsilon': 1e-6, 'clip': True, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142e-6, 1.4142e-6, 2., -2.]
    ), (
        {'epsilon': 1e-3, 'clip': True, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142e-3, 1.4142e-3, 2., -2.]
    ), (
        {'epsilon': 1., 'clip': True, 'norm_scale': True},
        [0., -0., 2., -2.],
        [1.4142, 1.4142, 2., -2.]
    ),
])
def test_stabilize(kwargs, input, expected):
    '''Test whether stabilize produces the expected outputs given some inputs.'''
    input_tensor = torch.tensor(input, dtype=torch.float64)
    output = stabilize(input_tensor, dim=0, **kwargs)
    expected_tensor = torch.tensor(expected, dtype=torch.float64)
    assert torch.allclose(expected_tensor, output)


@pytest.mark.parametrize('kwargs', prodict(
    epsilon=[1e-6, 1e-3, 1.],
    clip=[True, False],
    norm_scale=[True, False],
    dim=[None, (0,), (1,), (0, 1)]
))
def test_stabilizer_match(kwargs, data_simple):
    '''Test whether stabilize and Stabilizer produce the same output.'''
    stabilizer = Stabilizer(**kwargs)
    stabilizer_out = stabilizer(data_simple)
    stabilize_out = stabilize(data_simple, **kwargs)
    assert torch.allclose(stabilizer_out, stabilize_out)


@pytest.mark.parametrize('value', [1., 1, Stabilizer(epsilon=1.), lambda x: x])
def test_stabilizer_ensure(value):
    '''Test whether Stabilizer.ensure produces a stabilizer with the correct epsilon, or returns callables as-is.'''
    ensured = Stabilizer.ensure(value)
    assert not isinstance(value, float) or isinstance(ensured, Stabilizer) and ensured.epsilon == value
    assert not callable(value) or value is ensured


@pytest.mark.parametrize('value', [None, 'wow'])
def test_stabilizer_ensure_fail(value):
    '''Test whether Stabilizer.ensure fails on unsupported types.'''
    with pytest.raises(TypeError):
        Stabilizer.ensure(value)


@pytest.mark.parametrize('input_shape,target_shape,cut_batch_dim', [
    ((), (), False),
    ((), (2,), False),
    ((2,), (2, 2, 2), False),
    ((1, 2), (2, 2, 2), False),
    ((2, 1, 2), (2, 2, 2), None),
    ((2, 1, 2), (2, 2, 2), False),
    ((3, 1, 2), (2, 2, 2), True),
])
def test_expand_shapes(input_shape, target_shape, cut_batch_dim):
    '''Test whether expand produces correct shapes.'''
    kwargs = {} if cut_batch_dim is None else {'cut_batch_dim': cut_batch_dim}
    input = torch.zeros(input_shape)
    output = expand(input, target_shape, **kwargs)
    assert target_shape == output.shape


def test_expand_non_tensor():
    '''Test whether expand supports non-tensor input.'''
    target_shape = (2, 2, 2)
    input = [0., 1.]
    output = expand(input, target_shape)
    assert target_shape == output.shape


@pytest.mark.parametrize('input_shape,target_shape', [
    ((2,), (3,)),
    ((3,), (2, 2, 2)),
    ((3, 1, 2), (2, 2, 2)),
])
def test_expand_shape_invalid(input_shape, target_shape):
    '''Test whether expand raises RuntimeError on invalid shapes.'''
    input = torch.zeros(input_shape)
    with pytest.raises(RuntimeError):
        expand(input, target_shape)


@pytest.mark.parametrize('param_keys,require_params,zero_params,bias', [
    *product(
        [None, ['weight'], ['bias'], ['weight', 'bias']],
        [True],
        [[], ['weight'], ['bias'], ['weight', 'bias']],
        [False, True]
    ),
    *product([['weight', 'bias', 'beight', 'wias']], [False], [[]], [False, True]),
])
def test_param_mod(param_keys, require_params, zero_params, bias):
    '''Test whether ParamMod correctly changes and restores parameters of a module.'''
    module = nograd(torch.nn.Linear(2, 2, bias=bias))
    mod_param_keys = param_keys
    if param_keys is None:
        param_keys = [name for name, _ in module.named_parameters(recurse=False)]

    for key in param_keys:
        if getattr(module, key, None) is not None:
            getattr(module, key).data = torch.full_like(getattr(module, key), 0.5)

    param_mod = ParamMod(
        lambda x, _: torch.ones_like(x),
        param_keys=mod_param_keys,
        require_params=require_params,
        zero_params=zero_params,
    )

    with param_mod(module) as modified:
        for key in param_keys:
            if getattr(module, key, None) is not None:
                if key not in zero_params:
                    assert torch.all(getattr(modified, key) == 1.0), f'Parameter \'{key}\' was not modified!'
                else:
                    assert torch.all(getattr(modified, key) == 0.0), f'Parameter \'{key}\' was not set to zero!'

    for key in param_keys:
        if getattr(module, key, None) is not None:
            assert torch.all(getattr(module, key) == 0.5), f'Parameter \'{key}\' was not restored!'


def test_param_mod_required_missing():
    '''Test wether ParamMod raises a RuntimeError when it is missing parameters when require_params=True.'''
    module = torch.nn.Module()
    param_keys = ['beight', 'wias']
    param_mod = ParamMod(lambda x, _: torch.ones_like(x), param_keys=param_keys, require_params=True)
    with pytest.raises(RuntimeError):
        with param_mod(module):
            pass


@pytest.mark.parametrize('modifier', [lambda x: x, ParamMod(lambda x: x)])
def test_param_mod_ensure(modifier):
    '''Test wether ParamMod.ensure returns the original object if it is an instance of ParamMod, or a new instance of
    ParamMod if it is not an instance of ParamMod and callable.'''
    result = ParamMod.ensure(modifier)
    assert isinstance(result, ParamMod)
    assert not isinstance(modifier, ParamMod) or result is modifier


def test_param_mod_ensure_unknown_type():
    '''Test wether ParamMod.ensure raises a TypeError when it is called on a non-callable object that is not of type
    ParamMod.'''
    with pytest.raises(TypeError):
        ParamMod.ensure(None)


def test_collect_leaves_dummy():
    '''Test whether collect leaves correctly collects all leaves of an example model.'''
    modules = []

    def add(module):
        modules.append(module)
        return module

    class DummyModel(torch.nn.Module):
        '''Dummy model to produce a hierarchy of modules.'''
        def __init__(self):
            super().__init__()
            self.linear = add(torch.nn.Linear(2, 2))
            self.relu1 = add(torch.nn.ReLU())
            self.more = torch.nn.Sequential(
                add(torch.nn.Linear(2, 2)),
                add(torch.nn.ReLU()),
                add(torch.nn.Linear(2, 2)),
                add(torch.nn.ReLU()),
                add(torch.nn.Linear(2, 2)),
            )

    model = DummyModel()
    leaves = collect_leaves(model)
    assert all(a is b for a, b in zip(modules, leaves))


def test_collect_leaves_no_children():
    '''Test whether collect leaves correctly yields the input if it has no children.'''
    module = torch.nn.Module()
    leaf, = collect_leaves(module)
    assert leaf is module


def test_identity():
    '''Test whether Identity gives the correct output and grad and produces a grad_fn'''
    data = torch.randn(5, requires_grad=True)
    output, = Identity.apply(data)
    assert torch.allclose(data, output)
    assert hasattr(output, 'grad_fn')

    grad, = torch.autograd.grad(output, data, data)
    assert torch.allclose(data, grad)


def test_hook_grad():
    '''Test whether the Hook is correctly called in the forward and gradient passes.'''
    called = set()

    linear = torch.nn.Linear(2, 2)
    data = torch.randn(1, 2, requires_grad=True)
    grad_out = torch.ones(1, 2)

    class DummyHook(Hook):
        '''Dummy subclass of Hook to check whether forward and backward are called correctly.'''
        def forward(self, module, input, output):
            '''Check whether forward is called, and if the arguments are as expected.'''
            called.add('forward')
            assert module is linear
            assert torch.allclose(data, input[0])
            assert data is not input

        def backward(self, module, grad_input, grad_output):
            '''Check whether backward is called, and if the arguments are as expected.'''
            called.add('backward')
            assert grad_output[0] is grad_out

    hook = DummyHook()
    handles = hook.register(linear)
    try:
        out = linear(data)
        assert 'forward' in called
        assert 'backward' not in called
        torch.autograd.grad(out, data, grad_out)
        assert 'backward' in called
        assert 'grad_output' in hook.stored_tensors
        assert hook.stored_tensors['grad_output'][0] is grad_out
    finally:
        handles.remove()


def test_hook_no_grad():
    '''Test whether Hook's backward is never called when no gradient is required.'''
    called = set()

    data = torch.randn(1, 2, requires_grad=False)
    linear = torch.nn.Linear(2, 2)

    class DummyHook(Hook):
        '''Dummy subclass of Hook to check whether forward and backward are called.'''
        def forward(self, module, input, output):
            '''Check whether forward is called.'''
            called.add('forward')

        def backward(self, module, grad_input, grad_output):
            '''Check whether backward is called.'''
            called.add('backward')

    hook = DummyHook()
    handles = hook.register(linear)
    try:
        linear(data)
        assert 'forward' in called
        assert 'backward' not in called
    finally:
        handles.remove()


def test_hook_active():
    '''Test whether Hook's backward is only called when ``hook.active=True``.'''
    called = set()

    linear = torch.nn.Linear(2, 2)
    data = torch.randn(1, 2, requires_grad=True)

    class DummyHook(Hook):
        '''Dummy subclass of Hook to check whether forward and backward are called.'''
        def forward(self, module, input, output):
            '''Check whether forward is called.'''
            called.add('forward')

        def backward(self, module, grad_input, grad_output):
            '''Check whether backward is called.'''
            called.add('backward')

    hook = DummyHook()
    for active in (False, True):
        handles = hook.register(linear)
        try:
            hook.active = active
            out = linear(data)
            torch.autograd.grad(out.sum(), data)
            assert 'forward' in called
            assert active == ('backward' in called)
        finally:
            handles.remove()


def test_hook_grad_grad():
    '''Test whether the second order gradient is correct when registering a Hook.'''
    linear = torch.nn.Linear(2, 2)
    data = torch.randn(1, 2, requires_grad=True)

    class InputXGrad(Hook):
        '''InputXGrad implementation for testing.'''
        def forward(self, module, input, output):
            '''Store input for backward pass'''
            self.stored_tensors['input'] = input

        def backward(self, module, grad_input, grad_output):
            '''Compute InputXGrad.'''
            hin = self.stored_tensors['input'][0]
            with torch.autograd.enable_grad():
                hout = module.forward(hin)
            hgrad, = torch.autograd.grad(hout, hin, grad_outputs=grad_output[0], create_graph=True)
            return (hin * hgrad,)

    # true (modified) grad
    true_grad = data * linear.weight.sum(0)
    true_gradgrad = linear.weight.sum(0)

    hook = InputXGrad()
    handles = hook.register(linear)
    try:
        out = linear(data)

        # grad should be the true (modified) grad
        grad, = torch.autograd.grad(out, data, torch.ones_like(out), create_graph=True)
        assert torch.allclose(grad, true_grad)

        # wrong grad computation, will be equal to grad
        gradgrad, = torch.autograd.grad(grad, data, torch.ones_like(grad), retain_graph=True)
        assert torch.allclose(gradgrad, grad)

        hook.active = False
        # setting hook.active=False should now give the correct gradgrad
        gradgrad, = torch.autograd.grad(grad, data, torch.ones_like(grad), retain_graph=True)
        assert torch.allclose(gradgrad, true_gradgrad)
    finally:
        hook.active = True
        handles.remove()
    # removing the hook should also give the correct gradgrad
    gradgrad, = torch.autograd.grad(grad, data, torch.ones_like(grad))
    assert torch.allclose(gradgrad, true_gradgrad)


def test_hook_copy():
    '''Test whether copying a subclass of Hook creates an instance of the appropriate type.'''

    class DummyHook(Hook):
        '''Dummy subclass of Hook to check for the same class when copying.'''

    original = DummyHook()
    copy = original.copy()
    assert original is not copy
    assert isinstance(copy, DummyHook)


def test_hook_tuple():
    '''Test whether pre- and post_forward handle non-tuple inputs and tuple outputs.'''
    hook = Hook()
    module = None
    data = torch.ones(2)

    pre_out = hook.pre_forward(module, data)
    assert pre_out is data
    post_out = hook.post_forward(module, data, (data,))
    assert post_out is data


def test_basic_hook_default():
    '''Check whether BasicHook with default parameters runs as expected.'''
    linear = torch.nn.Linear(2, 2)
    data = torch.randn(1, 2, requires_grad=True)
    grad_out = torch.ones(1, 2)

    hook = BasicHook()
    handles = hook.register(linear)
    try:
        out = linear(data)
        assert 'input' in hook.stored_tensors
        assert hook.stored_tensors['input'][0] is not data
        assert torch.allclose(hook.stored_tensors['input'][0], data)
        torch.autograd.grad(out, data, grad_out)
        assert 'grad_output' in hook.stored_tensors
        assert hook.stored_tensors['grad_output'][0] is grad_out
    finally:
        handles.remove()


def test_basic_hook_custom():
    '''Check whether BasicHook with custom parameters runs as expected.'''
    linear = torch.nn.Linear(2, 2, bias=True)
    data = torch.randn(1, 2, requires_grad=True)
    grad_out = torch.ones(1, 2)

    called = set()

    def assert_params(obj, name=None):
        '''Assert whether param modifier function is called as expected.'''
        called.add('params')
        if name == 'weight':
            assert torch.allclose(obj, linear.weight.data)
        elif name == 'bias':
            assert torch.allclose(obj, linear.bias.data)
        return obj

    def assert_inputs(obj):
        '''Assert whether input modifier function is called as expected.'''
        called.add('input')
        assert torch.allclose(obj, data)
        return obj

    def assert_outputs(obj):
        '''Assert whether output modifier function is called as expected.'''
        called.add('output')
        return obj

    def assert_gradient(out_grad, outputs):
        '''Assert whether gradient mapper function is called as expected.'''
        called.add('output')
        called.add('gradient')
        assert out_grad is grad_out
        return [out_grad for output in outputs]

    def assert_reducer(inputs, gradients):
        '''Assert whether reducer function is called as expected.'''
        called.add('reducer')
        assert len(inputs) == len(gradients)
        assert torch.allclose(inputs[0], data)
        return sum(input * gradient for input, gradient in zip(inputs, gradients))

    hook = BasicHook(
        input_modifiers=[assert_inputs],
        param_modifiers=[
            ParamMod(
                assert_params,
                param_keys=['weight', 'bias'],
                require_params=True,
            )
        ],
        output_modifiers=[assert_outputs],
        gradient_mapper=assert_gradient,
        reducer=assert_reducer,
    )

    call_events = ('input', 'params', 'output', 'gradient', 'reducer')
    handles = hook.register(linear)
    try:
        out = linear(data)
        assert 'input' in hook.stored_tensors
        assert hook.stored_tensors['input'][0] is not data
        assert torch.allclose(hook.stored_tensors['input'][0], data)
        for key in call_events:
            assert key not in called, f'Event {key} was prematurely called in hook!'
        torch.autograd.grad(out, data, grad_out)
        assert 'grad_output' in hook.stored_tensors
        assert hook.stored_tensors['grad_output'][0] is grad_out
        for key in call_events:
            assert key in called, f'Event {key} was not called in hook!'
    finally:
        handles.remove()


def test_basic_hook_copy():
    '''Test whether BasicHook.copy copies the Hook correctly.'''
    hook = BasicHook(
        input_modifiers=[lambda obj, name: obj],
        param_modifiers=[
            ParamMod(
                (lambda obj, name: obj),
                param_keys=['weight', 'bias'],
                require_params=True,
                zero_params=['bias'],
            ),
        ],
        output_modifiers=[lambda obj, name: obj],
        gradient_mapper=(lambda out_grad, outputs: [out_grad for output in outputs]),
        reducer=(lambda inputs, gradients: [input * gradient for input, gradient in zip(inputs, gradients)]),
    )
    copy = hook.copy()

    attributes = (
        'input_modifiers',
        'param_modifiers',
        'output_modifiers',
        'reducer',
        'gradient_mapper',
    )

    assert copy is not hook
    for key in attributes:
        assert getattr(copy, key) is getattr(hook, key)


def test_removable_handle_deleted():
    '''Test whether RemovableHandle does nothing when attempting to call remove on a garbage collected object.'''

    called = set()

    class Dummy:
        '''Dummy object with a remove function which can be used for a RemovableHandle.'''
        @staticmethod
        def remove():
            '''Dummy remove function which tracks whether it has been called.'''
            called.add('remove')

    obj = Dummy()
    handle = RemovableHandle(obj)
    del obj
    handle.remove()
    assert 'remove' not in called


def test_composite_context():
    '''Test whether the composite context correctly uses the module map.'''
    linear = torch.nn.Linear(2, 2)
    called = set()

    class DummyHandle:
        '''Dummy handle to track whether it has been removed.'''
        @staticmethod
        def remove():
            '''Track whether this functions has been called.'''
            called.add('remove')

    handle = DummyHandle()

    class DummyHook:
        '''Dummy hook to track whether it has been registered correctly, and copied.'''
        @staticmethod
        def register(child):
            '''Assert whether child is the expected module and track that it has been called.'''
            assert child is linear
            called.add('register')
            return handle

        def copy(self):
            '''Return self an track that it has been called.'''
            called.add('copy')
            return self

    def module_map(ctx, name, child):
        '''Module map function to check whether all arguments are as expected which returns a DummyHook.'''
        assert child is linear
        assert name == ''
        assert isinstance(ctx, dict)
        return DummyHook()

    composite = Composite(module_map=module_map)

    with composite.context(linear):
        assert isinstance(composite.handles, RemovableHandleList)
        assert handle in composite.handles
        assert 'register' in called
        assert 'copy' in called
        assert 'remove' not in called
    assert 'remove' in called


def test_composite_empty():
    '''Test whether an empty composite with canonizers uses the canonizers as expected.'''
    linear = torch.nn.Linear(2, 2)
    called = set()

    class DummyHandle:
        '''Dummy handle to track whether its remove function has been called.'''
        @staticmethod
        def remove():
            '''Track whether the remove function was called.'''
            called.add('remove')

    handle = DummyHandle()

    class DummyCanonizer:
        '''Dummy canonizer to track whether its apply function has been called.'''
        @staticmethod
        def apply(module):
            '''Track whether the apply function has been called and return a dummy handle.'''
            called.add('apply')
            return [handle]

    canonizer = DummyCanonizer()
    composite = Composite(canonizers=[canonizer])

    with composite.context(linear):
        assert handle in composite.handles
        assert 'apply' in called
        assert 'remove' not in called
    assert 'remove' in called


def test_composite_inactive():
    '''Test whether a composite changes its hook's ``.active`` as expected.'''
    linear = torch.nn.Linear(2, 2)

    def module_map(ctx, name, module):
        return Hook()

    composite = Composite(module_map=module_map)

    # ensure no hooks are registered
    assert not list(islice(composite.hook_refs, 1))
    with composite.context(linear):
        # ensure hooks are registered
        assert list(islice(composite.hook_refs, 1))
        # check if all hooks are active
        assert all(hook.active for hook in composite.hook_refs)
        with composite.inactive():
            # check if no hooks are active
            assert all(not hook.active for hook in composite.hook_refs)
        # check that all hooks are active again
        assert all(hook.active for hook in composite.hook_refs)
    # ensure no hooks are registered
    assert not list(islice(composite.hook_refs, 1))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import sys
import os
from subprocess import run, CalledProcessError
import inspect
import pkg_resources

from pybtex.style.formatting.plain import Style as PlainStyle
from pybtex.style.labels import BaseLabelStyle
from pybtex.plugin import register_plugin


# -- Project information -----------------------------------------------------
project = 'zennit'
copyright = '2021, chr5tphr'
author = 'chr5tphr'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.linkcode',
    'sphinx.ext.mathjax',
    'sphinx.ext.extlinks',
    'sphinx_rtd_theme',
    'sphinx_copybutton',
    'sphinxcontrib.datatemplates',
    'sphinxcontrib.bibtex',
    'nbsphinx',
]


def config_inited_handler(app, config):
    os.makedirs(os.path.join(app.srcdir, app.config.generated_path), exist_ok=True)


def setup(app):
    app.add_config_value('REVISION', 'master', 'env')
    app.add_config_value('generated_path', '_generated', 'env')
    app.connect('config-inited', config_inited_handler)


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# interactive badges for binder and colab
nbsphinx_prolog = r"""
{% set docname = 'docs/source/' + env.doc2path(env.docname, base=False) %}

.. raw:: html

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/chr5tphr/zennit/blob/{{ env.config.REVISION }}/{{ docname|e }}">{{ docname|e }}</a>
      <br />
      Interactive online version:
      <span style="white-space: nowrap;">
        <a href="https://mybinder.org/v2/gh/chr5tphr/zennit/{{ env.config.REVISION|e }}?filepath={{ docname|e }}">
            <img alt="launch binder" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom">
        </a>
      </span>
      <span style="white-space: nowrap;">
        <a href="https://colab.research.google.com/github/chr5tphr/zennit/blob/{{ env.config.REVISION|e }}/{{ docname|e }}">
            <img alt="Open in Colab" src="https://colab.research.google.com/assets/colab-badge.svg" style="vertical-align:text-bottom">
        </a>
      </span>
    </div>
"""

# autosummary_generate = True

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"
copybutton_here_doc_delimiter = "EOT"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
# html_theme = 'alabaster'

html_favicon = '_static/favicon.svg'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

bibtex_bibfiles = ['bibliography.bib']
bibtex_default_style = 'author_year_style'
bibtex_reference_style = 'author_year'


class AuthorYearLabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        for entry in sorted_entries:
            yield f'[{entry.persons["author"][0].last_names[0]} et al., {entry.fields["year"]}]'


class AuthorYearStyle(PlainStyle):
    default_label_style = AuthorYearLabelStyle


register_plugin('pybtex.style.formatting', 'author_year_style', AuthorYearStyle)


def getrev():
    try:
        revision = run(
            ['git', 'describe', '--tags', 'HEAD'],
            capture_output=True,
            check=True,
            text=True
        ).stdout[:-1]
    except CalledProcessError:
        revision = 'master'

    return revision


REVISION = getrev()

extlinks = {
    'repo': (
        f'https://github.com/chr5tphr/zennit/blob/{REVISION}/%s',
        '%s'
    )
}

LINKCODE_URL = (
    f'https://github.com/chr5tphr/zennit/blob/{REVISION}'
    '/src/{filepath}#L{linestart}-L{linestop}'
)


# revised from https://gist.github.com/nlgranger/55ff2e7ff10c280731348a16d569cb73
def linkcode_resolve(domain, info):
    if domain != 'py' or not info['module']:
        return None

    modname = info['module']
    topmodulename = modname.split('.')[0]
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    try:
        modpath = pkg_resources.require(topmodulename)[0].location
        filepath = os.path.relpath(inspect.getsourcefile(obj), modpath)
        if filepath is None:
            return
    except Exception:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        return None
    else:
        linestart, linestop = lineno, lineno + len(source) - 1

    return LINKCODE_URL.format(filepath=filepath, linestart=linestart, linestop=linestop)



'''Zennit top-level __init__.'''
from . import attribution
from . import canonizers
from . import cmap
from . import composites
from . import core
from . import image
from . import layer
from . import rules
from . import torchvision
from . import types


__all__ = [
    'attribution',
    'canonizers',
    'cmap',
    'composites',
    'core',
    'image',
    'layer',
    'rules',
    'torchvision',
    'types',
]
